cimport cython
from cpython.list cimport PyList_Append, PyList_Size, PyList_GetItem


@cython.freelist(1000000)
cdef class MassObject(object):

    def __init__(self, obj, mass):
        self.obj = obj
        self.mass = mass

    def __repr__(self):
        return "MassObject(%r, %r)" % (self.obj, self.mass)

    @classmethod
    def from_chromatogram(cls, obj):
        return cls(obj, obj.weighted_neutral_mass)


@cython.boundscheck(False)
@cython.cdivision(True)
cpdef (ssize_t, bint) binary_search_with_flag(list array, double mass, double error_tolerance=1e-5):
    """Binary search an ordered array of objects with :attr:`neutral_mass`
    using a PPM error tolerance of `error_toler

    Parameters
    ----------
    array : list
        An list of objects, sorted over :attr:`neutral_mass` in increasing order
    mass : float
        The mass to search for
    error_tolerance : float, optional
        The PPM error tolerance to use when deciding whether a match has been found

    Returns
    -------
    int:
        The index in `array` of the best match
    bool:
        Whether or not a match was actually found, used to
        signal behavior to the caller.
    """
    cdef:
        ssize_t lo, hi, n, mid, best_index, i
        double err, best_error
        MassObject x
    lo = 0
    n = hi = len(array)
    while hi != lo:
        mid = (hi + lo) // 2
        x = <MassObject>PyList_GetItem(array, mid)
        err = (x.mass - mass) / mass
        if abs(err) <= error_tolerance:
            best_index = mid
            best_error = abs(err)
            i = mid - 1
            while i >= 0:
                x = <MassObject>PyList_GetItem(array, i)
                err = abs((x.mass - mass) / mass)
                if err < best_error:
                    best_error = err
                    best_index = i
                elif err > error_tolerance:
                    break
                i -= 1

            i = mid + 1
            while i < n:
                x = <MassObject>PyList_GetItem(array, i)
                err = abs((x.mass - mass) / mass)
                if err < best_error:
                    best_error = err
                    best_index = i
                elif err > error_tolerance:
                    break
                i += 1
            return best_index, True
        elif (hi - lo) == 1:
            return mid, False
        elif err > 0:
            hi = mid
        elif err < 0:
            lo = mid
    return 0, False


cdef class ChromatogramIndex(object):

    def __init__(self, chromatograms, sort=True):
        if sort:
            self._sort(chromatograms)
        else:
            self._chromatograms = list(map(MassObject.from_chromatogram, chromatograms))

    def _sort(self, iterable):
        self._chromatograms = [
            MassObject.from_chromatogram(c) for c in sorted(
                [c for c in iterable if len(c)], key=lambda x: (x.neutral_mass, x.start_time))]

    def add(self, chromatogram, sort=True):
        self._chromatograms.append(MassObject.from_chromatogram(chromatogram))
        if sort:
            self._sort(self.chromatograms)

    def __len__(self):
        return self.get_size()

    def __getitem__(self, i):
        cdef:
            MassObject obj
            slice si
            size_t j, n
            list slice_out, slice_in
        if isinstance(i, slice):
            si = <slice>i
            slice_out = []
            slice_in = self._chromatograms[si]
            n = PyList_Size(slice_in)
            for j in range(n):
                PyList_Append(slice_out, (<MassObject>PyList_GetItem(slice_in, j)).obj)
            return slice_out
        obj = <MassObject>PyList_GetItem(self._chromatograms, i)
        return obj.obj

    cdef inline MassObject get_index(self, size_t i):
        obj = <MassObject>PyList_GetItem(self._chromatograms, i)
        return obj

    cdef inline size_t get_size(self):
        return PyList_Size(self._chromatograms)

    def __iter__(self):
        cdef:
            size_t i, n
            MassObject obj
        n = self.get_size()
        for i in range(n):
            obj = <MassObject>PyList_GetItem(self._chromatograms, i)
            yield obj.obj

    @property
    def chromatograms(self):
        return self._chromatograms

    @chromatograms.setter
    def chromatograms(self, value):
        self._chromatograms = list(map(MassObject.from_chromatogram, value))

    cpdef (ssize_t, bint) _binary_search(self, double mass, double error_tolerance=1e-5):
        return binary_search_with_flag(self._chromatograms, mass, error_tolerance)

    cpdef object find_mass(self, double mass, double ppm_error_tolerance=1e-5):
        index, flag = self._binary_search(mass, ppm_error_tolerance)
        if flag:
            return self[index]
        else:
            return None

    @cython.cdivision(True)
    @cython.boundscheck(False)
    cpdef object find_all_by_mass(self, double mass, double ppm_error_tolerance=1e-5):
        cdef:
            ssize_t i, n, center_index, low_index, high_index
            bint flag
            list items
            MassObject x

        n = self.get_size()
        if n == 0:
            return self.__class__([], sort=False)
        center_index, flag = self._binary_search(mass, ppm_error_tolerance)
        low_index = center_index
        while low_index > 0:
            x = self.get_index(low_index - 1)
            if abs((mass - x.mass) / x.mass) > ppm_error_tolerance:
                break
            low_index -= 1

        high_index = center_index - 1
        while high_index < n - 1:
            x = self.get_index(high_index + 1)
            if abs((mass - x.mass) / x.mass) > ppm_error_tolerance:
                break
            high_index += 1

        if low_index == high_index == center_index:
            x = self.get_index(center_index)
            if abs((mass - x.mass) / x.mass) > ppm_error_tolerance:
                return self.__class__([], sort=False)

        items = []
        for i in range(low_index, high_index + 1):
            x = self.get_index(i)
            if (abs(x.mass - mass) / mass) < ppm_error_tolerance:
                PyList_Append(items, x.obj)

        return self.__class__(items, sort=False)

    cpdef object mass_between(self, double low, double high):
        cdef:
            size_t n
            ssize_t low_index, high_index
        n = self.get_size()
        if n == 0:
            return self.__class__([])
        low_index, flag = self._binary_search(low, 1e-5)
        low_index = max(0, min(low_index, n - 1))
        if self.get_index(low_index).mass  < low:
            low_index += 1
        high_index, flag = self._binary_search(high, 1e-5)
        high_index += 2
        # high_index = min(n - 1, high_index)
        if (high_index < n) and (self.get_index(high_index).mass > high):
            high_index -= 1
        items = self[low_index:high_index]
        items = [c for c in items if low <= c.neutral_mass <= high]
        return self.__class__(items, sort=False)

    def __repr__(self):
        return repr(list(self))

    def _repr_pretty_(self, p, cycle):
        return p.pretty(list(self))

    def __str__(self):
        return str(list(self))



import math

import itertools
from collections import defaultdict
from typing import Any, DefaultDict, Dict, Iterator, List, Set, Tuple, TYPE_CHECKING, Optional

from glycresoft.chromatogram_tree.mass_shift import MassShiftBase

from ms_deisotope import DeconvolutedPeak
from ms_deisotope.data_source import ProcessedScan

try:
    from collections.abc import Sequence
except ImportError:
    from collections import Sequence

from glypy.structure.glycan_composition import FrozenMonosaccharideResidue, HashableGlycanComposition

from glycresoft.chromatogram_tree import Unmodified

from .fragment_match_map import SpectrumGraph, PeakPairTransition

if TYPE_CHECKING:
    from glycresoft.database.mass_collection import NeutralMassDatabase

hexnac = FrozenMonosaccharideResidue.from_iupac_lite("HexNAc")
hexose = FrozenMonosaccharideResidue.from_iupac_lite("Hex")
xylose = FrozenMonosaccharideResidue.from_iupac_lite("Xyl")
fucose = FrozenMonosaccharideResidue.from_iupac_lite("Fuc")
neuac = FrozenMonosaccharideResidue.from_iupac_lite("NeuAc")
neugc = FrozenMonosaccharideResidue.from_iupac_lite("NeuGc")


def collect_glycan_composition_from_annotations(path: 'Path') -> HashableGlycanComposition:
    gc = HashableGlycanComposition()
    for edge in path:
        if isinstance(edge.annotation, tuple):
            for part in edge.annotation:
                gc[part] += 1
        else:
            gc[edge.annotation] += 1
    if path.end_node is not None:
        if isinstance(path.end_node, tuple):
            for part in path.end_node:
                gc[part] += 1
        else:
            gc[path.end_node] += 1
    return gc


class MassWrapper(object):
    '''An adapter class to make types whose mass calculation is a method
    (:mod:`glypy` dynamic graph components) compatible with code where the
    mass calculation is an  attribute (:mod:`glycopeptidepy` objects and
    most things here)

    Hashes and compares as :attr:`obj`

    Attributes
    ----------
    obj: object
        The wrapped object
    mass: float
        The mass of :attr:`obj`
    '''
    obj: Any
    mass: float

    def __init__(self, obj, mass=None):
        self.obj = obj
        if mass is not None:
            self.mass = mass
        else:
            try:
                # object's mass is a method
                self.mass = obj.mass()
            except TypeError:
                # object's mass is a plain attribute
                self.mass = obj.mass

    def __repr__(self):
        if isinstance(self.obj, tuple):
            filling = ', '.join(tuple(map(str, self.obj)))
        else:
            filling = str(self.obj)

        return f"{self.__class__.__name__}({filling})"

    def __eq__(self, other):
        return self.obj == other

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self.obj)

    def __lt__(self, other):
        return self.mass < other.mass

    def __gt__(self, other):
        return self.mass > other.mass


default_components = (hexnac, hexose, xylose, fucose, neuac, neugc)


class PeakGroup(object):
    """
    An adapter for a collection of :class:`~.DeconvolutedPeak` objects which share
    the same approximate neutral mass that looks like a single :class:`~.DeconvolutedPeak`
    object for edge-like calculations with :class:`EdgeGroup`.

    Attributes
    ----------
    peaks: tuple
        A tuple of the *unique* peaks forming this group
    neutral_mass: float
        The intensity weighted average neutral mass of the peaks in this group
    intensity: float
        The sum of the intensities of the peaks in this group
    """
    peaks: Tuple[DeconvolutedPeak]
    neutral_mass: float
    intensity: float

    def __init__(self, peaks):
        self.peaks = tuple(sorted(set(peaks), key=lambda x: (x.neutral_mass, x.charge)))
        self.neutral_mass = 0.0
        self.intensity = 0.0
        self._hash = hash((p.index.neutral_mass for p in self.peaks))
        self._initialize()

    def _initialize(self):
        neutral_mass_acc = 0.0
        intensity = 0.0
        for peak in self.peaks:
            neutral_mass_acc += peak.neutral_mass * peak.intensity
            intensity += peak.intensity
        self.intensity = intensity
        self.neutral_mass = neutral_mass_acc / intensity

    def __eq__(self, other):
        return self.peaks == other.peaks

    def __hash__(self):
        return self._hash

    def __getitem__(self, i: int) -> DeconvolutedPeak:
        return self.peaks[i]

    def __len__(self):
        return len(self.peaks)

    def __repr__(self):
        template = "{self.__class__.__name__}({self.neutral_mass}, {self.intensity}, {size})"
        return template.format(self=self, size=len(self))


class EdgeGroup(object):
    """
    An adapter for a collection of :class:`~.PeakPairTransition` objects which all belong to
    different :class:`Path` objects that share the same annotation sequence and approximate start
    and end mass.

    Attributes
    ----------
    start: PeakGroup
        The peaks that make up the starting mass for this edge group
    end: PeakGroup
        The peaks that make up the ending mass for this edge group
    annotation: object
        The common annotation for members of this group
    """

    start: PeakGroup
    annotation: Any
    end: PeakGroup

    _transitions: List[PeakPairTransition]
    _hash: int

    def __init__(self, transitions):
        self._transitions = transitions
        self.start = None
        self.end = None
        self._hash = None
        self.annotation = self._transitions[0].annotation
        self._initialize_by_transitions()

    def __hash__(self):
        return self._hash

    def __eq__(self, other: 'EdgeGroup'):
        if other is None:
            return False
        return self.start == other.start and self.end == other.end and self.annotation == other.annotation

    def __ne__(self, other):
        return not self == other

    def __getitem__(self, i):
        return self._transitions[i]

    def __len__(self):
        return len(self._transitions)

    def _initialize_by_transitions(self):
        if self._transitions:
            starts = [edge.start for edge in self._transitions]
            ends = [edge.end for edge in self._transitions]
            self.start = PeakGroup(starts)
            self.end = PeakGroup(ends)
            self._hash = hash((self.start.peaks[0].index.neutral_mass,
                               self.end.peaks[0].index.neutral_mass))
        else:
            self._hash = hash(())

    @classmethod
    def from_combination(cls, start: PeakGroup, end: PeakGroup, annotation: MassWrapper):
        return cls([PeakPairTransition(start_peak, end_peak, annotation)
                    for start_peak, end_peak in itertools.product(start, end)])

    @classmethod
    def aggregate(cls, paths):
        edges = []
        n = len(paths)
        if n == 0:
            return []
        m = len(paths[0])
        for i in range(m):
            group = []
            for j in range(n):
                group.append(paths[j][i])
            edges.append(group)
        return paths[0].__class__([cls(g) for g in edges])

    def __repr__(self):
        return ("{self.__class__.__name__}({self.start.neutral_mass:0.3f} "
                "-{self.annotation}-> {self.end.neutral_mass:0.3f})").format(self=self)


def collect_paths(paths: List['Path'], error_tolerance: float=1e-5) -> List[List['Path']]:
    """
    Group together paths which share the same annotation sequence and approximate
    start and end masses.

    Parameters
    ----------
    paths: :class:`list`
        A list of :class:`Path` objects to group

    Returns
    -------
    :class:`list` of :class:`list` of :class:`Path` objects
    """
    groups = defaultdict(list)
    if not paths:
        return []
    for path in paths:
        key = tuple(e.annotation for e in path)
        groups[key].append(path)
    result = []
    for key, block_members in groups.items():
        block_members = sorted(block_members, key=lambda x: x.start_mass)
        current_path = block_members[0]
        members = [current_path]
        for path in block_members[1:]:
            if abs(current_path.start_mass - path.start_mass) / path.start_mass < error_tolerance:
                members.append(path)
            else:
                result.append(members)
                current_path = path
                members = [current_path]
        result.append(members)
    return result


def score_path(path: 'Path') -> float:
    if not path:
        return 0
    acc = math.log10(path.transitions[0].start.intensity)
    for edge in path.transitions:
        acc += math.log10(edge.end.intensity)
    return acc


class Path(object):
    """
    Represent a single contiguous sequence of :class:`~.PeakPairTransition` objects
    forming a sequence tag, or a path along the edges of a peak graph.

    Attributes
    ----------
    transitions: :class:`list` of :class:`~.PeakPairTransition`
        The edges connecting pairs of peaks.
    total_signal: float
        The total signal captured by this path
    start_mass: float
        The lowest mass among all peaks in the path
    end_mass: float
        The highest mass among all peaks in the path
    """

    transitions: List[PeakPairTransition]
    total_signal: float
    start_mass: float
    end_mass: float

    start_node: Optional[Any]
    end_node: Optional[Any]

    _peaks_used: Optional[Set[DeconvolutedPeak]]
    _edges_used: Optional[DefaultDict[Tuple[DeconvolutedPeak, DeconvolutedPeak], List[Any]]]

    def __init__(self, edge_list, start_node=None, end_node=None):
        self.transitions = edge_list
        self.total_signal = self._total_signal()
        self.start_mass = self[0].start.neutral_mass
        self.end_mass = self[-1].end.neutral_mass
        self.start_node = start_node
        self.end_node = end_node
        self._peaks_used = None
        self._edges_used = None

    def copy(self) -> 'Path':
        return self.__class__(self.transitions, self.start_node, self.end_node)

    @property
    def peaks(self):
        if self._peaks_used is None:
            self._peaks_used = self._build_peaks_set()
        return self._peaks_used

    def _build_peaks_set(self) -> PeakPairTransition:
        peaks = set()
        for edge in self:
            peaks.add(edge.end)
        peaks.add(self[0].start)
        return peaks

    def _build_edges_used(self) -> DefaultDict[Tuple[DeconvolutedPeak, DeconvolutedPeak], List[Any]]:
        mapping = defaultdict(list)
        for edge in self:
            mapping[edge.start, edge.end].append(edge)
        return mapping

    def __contains__(self, edge: PeakPairTransition):
        return self.has_edge(edge, True)

    def has_edge(self, edge: PeakPairTransition, match_annotation: bool=False) -> bool:
        if self._edges_used is None:
            self._edges_used = self._build_edges_used()
        key = (edge.start, edge.end)
        if key in self._edges_used:
            edges = self._edges_used[key]
            if match_annotation:
                for e in edges:
                    if e == edge:
                        return True
                return False
            else:
                for e in edges:
                    if e.key != edge.key:
                        continue
                    if e.start != edge.start:
                        continue
                    if e.end != edge.end:
                        continue
                    return True
                return False

    def has_peak(self, peak: DeconvolutedPeak) -> bool:
        if self._peaks_used is None:
            self._peaks_used = self._build_peaks_set()
        return peak in self._peaks_used

    def has_peaks(self, peaks_set: Set[DeconvolutedPeak]) -> bool:
        if self._peaks_used is None:
            self._peaks_used = self._build_peaks_set()
        return peaks_set & self._peaks_used

    def __iter__(self) -> Iterator[PeakPairTransition]:
        return iter(self.transitions)

    def __getitem__(self, i):
        return self.transitions[i]

    def __len__(self):
        return len(self.transitions)

    def _total_signal(self) -> float:
        total = 0.0
        for edge in self:
            total += edge.end.intensity
        total += self[0].start.intensity
        return total

    def __repr__(self):
        transitions = '->'.join(_format_annotation(e.annotation)
                                for e in self)
        if self.start_node is not None:
            transitions = f"{_format_annotation(self.start_node)}:-{transitions}"
        if self.end_node is not None:
            transitions = f"{transitions}-:{_format_annotation(self.end_node)}"
        return f"{self.__class__.__name__}({transitions}, {self.total_signal:0.4e}, {self.start_mass}, {self.end_mass})"


def _format_annotation(annot: Any):
    return '[%s]' % ', '.join(map(str, annot)) if isinstance(annot, tuple) else str(annot)


class PathSet(Sequence):
    def __init__(self, paths, ordered=False):
        self.paths = (sorted(paths, key=lambda x: x.start_mass)
                      if not ordered else paths)

    def __getitem__(self, i):
        return self.paths[i]

    def __len__(self):
        return len(self.paths)

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, str(self.paths)[1:-1])

    def _repr_pretty_(self, p, cycle):
        with p.group(9, "%s([" % self.__class__.__name__, "])"):
            for i, path in enumerate(self):
                if i:
                    p.text(",")
                    p.breakable()
                p.pretty(path)

    def is_path_disjoint(self, path):
        for p in self:
            if p.has_peaks(path.peaks):
                return False
        return True

    def threshold(self, topn=100):
        return self.__class__(sorted(self, key=lambda x: x.total_signal, reverse=True)[:topn])


class PathFinder(object):
    components: List[MassWrapper]
    product_error_tolerance: float
    merge_paths: bool
    precursor_error_tolerance: float
    gap_completion_lookup: Dict[int, 'NeutralMassDatabase[MassWrapper]']
    _gap_completion_mass_range: Dict[int, Tuple[float, float]]

    def __init__(self, components=None,
                 product_error_tolerance: float=1e-5,
                 merge_paths: bool=True,
                 precursor_error_tolerance: float=1e-5):
        if components is None:
            components = default_components
        self.components = list(map(MassWrapper, components))
        self.product_error_tolerance = product_error_tolerance
        self.merge_paths = merge_paths
        self.precursor_error_tolerance = precursor_error_tolerance
        self.gap_completion_lookup = {}
        self._gap_completion_mass_range = {}

    def build_graph(self, scan: ProcessedScan, mass_shift: MassShiftBase=Unmodified) -> SpectrumGraph:
        graph = SpectrumGraph()
        has_tandem_shift = abs(mass_shift.tandem_mass) > 0
        for peak in scan.deconvoluted_peak_set:
            for component in self.components:
                for other_peak in scan.deconvoluted_peak_set.all_peaks_for(
                        peak.neutral_mass + component.mass, self.product_error_tolerance):
                    graph.add(peak, other_peak, component.obj)
                if has_tandem_shift:
                    for other_peak in scan.deconvoluted_peak_set.all_peaks_for(
                            peak.neutral_mass + component.mass + mass_shift.tandem_mass,
                            self.product_error_tolerance):
                        graph.add(peak, other_peak, component.obj)
        return graph

    def paths_from_graph(self, graph: SpectrumGraph, limit: int=200) -> List[Path]:
        paths = []
        min_start_mass = min(c.mass for c in self.components) + 1
        for path in graph.longest_paths(limit=limit):
            path = Path(path)
            if path.start_mass < min_start_mass:
                continue
            paths.append(path)
        return paths

    def aggregate_paths(self, paths: List['Path']) -> DefaultDict[Tuple[Any], List['Path']]:
        groups = defaultdict(list)
        for path in paths:
            key = tuple(e.annotation for e in path)
            groups[key].append(path)
        return groups

    def collect_paths(self, paths: List['Path'], error_tolerance: float = 1e-5) -> List[List['Path']]:
        """
        Group together paths which share the same annotation sequence and approximate
        start and end masses.

        Parameters
        ----------
        paths: :class:`list`
            A list of :class:`Path` objects to group

        Returns
        -------
        :class:`list` of :class:`list` of :class:`Path` objects
        """
        if not paths:
            return []
        groups = self.aggregate_paths(paths)
        result = []
        for _key, block_members in groups.items():
            block_members = sorted(block_members, key=lambda x: x.start_mass)
            current_path = block_members[0]
            members = [current_path]
            for path in block_members[1:]:
                if abs(current_path.start_mass - path.start_mass) / path.start_mass < error_tolerance:
                    members.append(path)
                else:
                    result.append(members)
                    current_path = path
                    members = [current_path]
            result.append(members)
        result = [EdgeGroup.aggregate(block) for block in result]
        return result

    def paths(self, scan: ProcessedScan, mass_shift: MassShiftBase=Unmodified, limit: int=200) -> List[Path]:
        graph = self.build_graph(scan, mass_shift)
        paths = self.paths_from_graph(graph, limit)
        if self.merge_paths:
            paths = self.collect_paths(paths, self.precursor_error_tolerance)
        return paths

    def _fill_gap_size_bin(self, size: int):
        combos = []
        for items in  itertools.combinations_with_replacement(self.components, size):
            step = MassWrapper(tuple(v.obj for v in items), sum([v.mass for v in items]))
            combos.append(step)
        from glycresoft.database.mass_collection import NeutralMassDatabase
        return NeutralMassDatabase(combos, lambda x: x.mass)

    def gap_bin_entries(self, size: int) -> 'NeutralMassDatabase[MassWrapper]':
        if size not in self.gap_completion_lookup:
            self.gap_completion_lookup[size] = self._fill_gap_size_bin(size)
            masses = [m.mass for m in self.gap_completion_lookup[size]]
            self._gap_completion_mass_range[size] = (min(masses), max(masses))
        return self.gap_completion_lookup[size]

    def complete_paths(self, paths: List[Path], max_gap_size: int = 2) -> List[Path]:
        from glycresoft.database.mass_collection import NeutralMassDatabase
        by_front = NeutralMassDatabase(paths, lambda x: x.start_mass)
        by_end = NeutralMassDatabase(paths, lambda x: x.end_mass)
        max_gap_size = 2
        extensions = []

        for path in by_end:
            for i in range(1, max_gap_size + 1):
                spacers = self.gap_bin_entries(i)
                for candidate in spacers:
                    suffixes = by_front.search_mass_ppm(
                        path.end_mass + candidate.mass, self.product_error_tolerance)
                    for suffix in suffixes:
                        extensions.append(Path(
                                list(path) + [EdgeGroup.from_combination(
                                path[-1].end, suffix[0].start, candidate.obj)] + list(suffix),
                                path.start_node,
                                suffix.end_node,
                            ))
        return extensions + paths

    def complete_precursor(self, scan: ProcessedScan,
                           path: Path,
                           mass_shift: MassShiftBase=Unmodified,
                           max_gap_size: int=3) -> List[Path]:
        precursor_mass = scan.precursor_information.neutral_mass
        leftover_mass: float = precursor_mass - (path.end_mass + mass_shift.mass)
        candidates: List[MassWrapper] = []
        for i in range(1, max_gap_size + 1):
            candidates.extend(
                self.gap_bin_entries(i).search_mass(
                leftover_mass, error_tolerance=0.2))
        result = []
        for candidate in candidates:
            proposed_mass = candidate.mass + path.end_mass + mass_shift.mass
            if abs(proposed_mass - precursor_mass) / precursor_mass <= self.precursor_error_tolerance:
                completed = path.copy()
                completed.end_node = candidate.obj
                result.append(completed)
        if not result:
            result = [path]
        return result





import sys
import cmd
import csv
import threading
import code
import pprint

from queue import Queue, Empty

import click
import pkg_resources

from glypy.structure.glycan_composition import HashableGlycanComposition
from glycopeptidepy.io import fasta, uniprot
from glycopeptidepy.structure.residue import UnknownAminoAcidException

from .base import cli
from .validators import RelativeMassErrorParam, get_by_name_or_id

from glycresoft.serialize import (
    DatabaseBoundOperation, GlycanHypothesis, GlycopeptideHypothesis,
    SampleRun, Analysis, Protein, Peptide, Glycopeptide, GlycanClass,
    GlycanComposition, func, FileBlob)

from glycresoft.database import (
    GlycopeptideDiskBackedStructureDatabase,
    GlycanCompositionDiskBackedStructureDatabase)

from glycresoft.database.builder.glycopeptide.proteomics import mzid_proteome
from glycresoft.database.builder.glycopeptide.proteomics.uniprot import UniprotProteinDownloader


@cli.group("tools", short_help="Odds and ends to help inspect data and diagnose issues")
def tools():
    pass


def database_connection(fn):
    arg = click.argument("database-connection", doc_help=(
        "A connection URI for a database, or a path on the file system"),
        type=DatabaseBoundOperation)
    return arg(fn)


def hypothesis_identifier(hypothesis_type):
    def wrapper(fn):
        arg = click.argument("hypothesis-identifier", doc_help=(
            "The ID number or name of the %s hypothesis to use" % (hypothesis_type,)))
        return arg(fn)
    return wrapper


def analysis_identifier(fn):
    arg = click.argument("analysis-identifier", doc_help=(
        "The ID number or name of the analysis to use"))
    return arg(fn)


@tools.command("list", short_help='List names and ids of collections in the database')
@click.pass_context
@database_connection
def list_contents(context, database_connection):
    click.echo("Glycan Hypothesis")
    for hypothesis in database_connection.query(GlycanHypothesis):
        click.echo("\t%d\t%s\t%s" % (hypothesis.id, hypothesis.name, hypothesis.uuid))

    click.echo("\nGlycopeptide Hypothesis")
    for hypothesis in database_connection.query(GlycopeptideHypothesis):
        click.echo("\t%d\t%s\t%s\t%d" % (hypothesis.id, hypothesis.name, hypothesis.uuid,
                                         hypothesis.glycan_hypothesis.id))

    click.echo("\nSample Run")
    for run in database_connection.query(SampleRun):
        click.echo("\t%d\t%s\t%s" % (run.id, run.name, run.uuid))

    click.echo("\nAnalysis")
    for analysis in database_connection.query(Analysis):
        click.echo("\t%d\t%s\t%s" % (analysis.id, analysis.name, analysis.sample_run.name))

    click.echo("\nFile Blobs")
    for blob in database_connection.query(FileBlob):
        click.echo("\t%d\t%s" % (blob.id, blob.name))


@tools.command('mzid-proteins', short_help='Extract proteins from mzIdentML files')
@click.argument("mzid-file")
def list_protein_names(mzid_file):
    for name in mzid_proteome.protein_names(mzid_file):
        click.echo(name)


class SQLShellInterpreter(cmd.Cmd):
    prompt = '[sql-shell] '

    def __init__(self, session, *args, **kwargs):
        self.session = session
        cmd.Cmd.__init__(self, *args, **kwargs)

    def postcmd(self, stop, line):
        if line == "quit":
            return True
        return False

    def precmd(self, line):
        tokens = line.split(" ")
        tokens[0] = tokens[0].lower()
        return ' '.join(tokens)

    def _print_table(self, result):
        rows = list(result)
        if len(rows) == 0:
            return
        sizes = [0] * len(rows[0])
        titles = rows[0].keys()
        str_rows = []
        for row in rows:
            str_row = []
            for i, col in enumerate(row):
                col = repr(col)
                col_len = len(col)
                if sizes[i] < col_len:
                    sizes[i] = col_len
                str_row.append(col)
            str_rows.append(str_row)
        print(" | ".join(titles))
        for row in str_rows:
            print(' | '.join(row))

    def _to_csv(self, rows, fh):
        titles = rows[0].keys()
        writer = csv.writer(fh)
        writer.writerow(titles)
        writer.writerows(rows)

    def do_export(self, line):
        try:
            fname, line = line.rsplit(";", 1)
            if len(fname.strip()) == 0 or len(line.strip()) == 0:
                raise ValueError()
            try:
                result = self.session.execute("select " + line)
            except Exception as e:
                print(str(e))
                self.session.rollback()
                return
            try:
                rows = list(result)
                print("%d rows selected; Writing to %r" % (len(rows), fname))
                with open(fname, 'wb') as handle:
                    self._to_csv(rows, handle)
            except Exception as e:
                print(str(e))

        except ValueError:
            print("Could not determine the file name to export to.")
            print("Usage: export <filename>; <query>")

    def do_execsql(self, line):
        try:
            result = self.session.execute(line)
            self._print_table(result)

        except Exception as e:
            print(str(e))
            self.session.rollback()
        return False

    def do_explain(self, line):
        try:
            result = self.session.execute("explain " + line)
            self._print_table(result)

        except Exception as e:
            print(str(e))
            self.session.rollback()
        return False

    def do_select(self, line):
        try:
            result = self.session.execute("select " + line)
            self._print_table(result)

        except Exception as e:
            print(str(e))
            self.session.rollback()
        return False

    def do_quit(self, line):
        return True


@tools.command("sql-shell",
               short_help=("A minimal SQL command shell for running "
                           "diagnostics and exporting arbitrary data."))
@click.argument("database-connection")
@click.option("-s", "--script", default=None)
def sql_shell(database_connection, script=None):
    db = DatabaseBoundOperation(database_connection)
    session = db.session()  # pylint: disable=not-callable
    interpreter = SQLShellInterpreter(session)
    if script is None:
        interpreter.cmdloop()
    else:
        result = session.execute(script)
        interpreter._to_csv(list(result), sys.stdout)


@tools.command("validate-fasta", short_help="Validates a FASTA file, checking a few errors.")
@click.argument("path")
def validate_fasta(path):
    with open(path, "r") as handle:
        n_deflines = 0
        for line in handle:
            if line.startswith(">"):
                n_deflines += 1
    with open(path, 'r') as handle:
        n_entries = 0
        for entry in fasta.FastaFileParser(handle):
            n_entries += 1
    if n_entries != n_deflines:
        click.secho("%d\">\" prefixed lines found, but %d entries parsed" % (n_deflines, n_entries))
    else:
        click.echo("%d Protein Sequences" % (n_entries, ))
    n_glycoprots = 0
    o_glycoprots = 0
    with open(path, 'r') as handle:
        invalid_sequences = []
        for entry in fasta.FastaFileParser(handle):
            try:
                seq = fasta.ProteinSequence(entry['name'], entry['sequence'])
                n_glycoprots += bool(seq.n_glycan_sequon_sites)
                o_glycoprots += bool(seq.o_glycan_sequon_sites)
            except UnknownAminoAcidException as e:
                invalid_sequences.append((entry['name'], e))
    click.echo("Proteins with N-Glycosites: %d" % n_glycoprots)
    for name, error in invalid_sequences:
        click.secho("%s had %s" % (name, error), fg='yellow')


@tools.command("validate-glycan-text", short_help="Validates a text file of glycan compositions")
@click.argument("path")
def validate_glycan_text(path):
    from glycresoft.database.builder.glycan.glycan_source import TextFileGlycanCompositionLoader
    with open(path, 'r') as handle:
        loader = TextFileGlycanCompositionLoader(handle)
        n = 0
        glycan_classes = set()
        residues = set()
        unresolved = set()
        for line in loader:
            n += 1
            glycan_classes.update(line[1])
            glycan_composition = HashableGlycanComposition.parse(line[0])
            for residue in glycan_composition.keys():
                if residue.mass() == 0:
                    unresolved.add(residue)
                residues.add(residue)
        click.secho("%d glycan compositions" % (n,))
        click.secho("Residues:")
        for residue in residues:
            click.secho("\t%s - %f" % (str(residue), residue.mass()))
        if unresolved:
            click.secho("Unresolved Residues:", fg='yellow')
            click.secho("\n".join(str(r) for r in unresolved), fg='yellow')


def has_known_glycosylation(accession):
    try:
        prot = uniprot.get(accession)
        if "Glycoprotein" in prot.keywords:
            return True
        else:
            # for feature in prot.features:
            #     if isinstance(feature, uniprot.GlycosylationSite):
            #         return True
            pass
        return False
    except Exception:
        return False


@tools.command("known-glycoproteins", short_help=(
    'Checks UniProt to see if a list of proteins contains any known glycoproteins'))
@click.option("-i", "--file-path", help="Read input from a file instead of STDIN")
@click.option("-f", "--fasta-format", is_flag=True, help="Indicate input is in FASTA format")
@click.option("-o", "--output-path", help="Write output to a file instead of STDOUT")
def known_uniprot_glycoprotein(file_path=None, output_path=None, fasta_format=False):
    if file_path is not None:
        handle = open(file_path)
    else:
        handle = sys.stdin

    if fasta_format:
        reader = fasta.ProteinFastaFileParser(handle)

        def name_getter(x):
            return x.name.accession
    else:
        reader = handle

        def name_getter(x):
            header = fasta.default_parser(x)
            try:
                return header.accession
            except Exception:
                return header[0]

    if output_path is None:
        outhandle = sys.stdout
    else:
        outhandle = open(output_path, 'w')

    def checker_task(inqueue, outqueue, no_more_event):
        has_work = True
        while has_work:
            try:
                protein = inqueue.get(True, 1)
            except Empty:
                if no_more_event.is_set():
                    has_work = False
                continue
            try:
                if has_known_glycosylation(name_getter(protein)):
                    outqueue.put(protein)
            except Exception as e:
                click.secho("%r occurred for %s" % (e, protein), err=True, fg='yellow')

    def consumer_task(outqueue, no_more_event):
        has_work = True
        if fasta_format:
            writer = fasta.ProteinFastaFileWriter(outhandle)
            write_fn = writer.write
        else:
            def write_fn(payload):
                outhandle.write(str(payload).strip() + '\n')
        while has_work:
            try:
                protein = outqueue.get(True, 1)
            except Empty:
                if no_more_event.is_set():
                    has_work = False
                continue
            write_fn(protein)
        outhandle.close()

    producer_done = threading.Event()
    checker_done = threading.Event()

    inqueue = Queue()
    outqueue = Queue()

    n_threads = 10
    checkers = [threading.Thread(
        target=checker_task, args=(inqueue, outqueue, producer_done)) for i in range(n_threads)]
    for check in checkers:
        check.start()

    consumer = threading.Thread(target=consumer_task, args=(outqueue, checker_done))
    consumer.start()

    for protein in reader:
        inqueue.put(protein)

    producer_done.set()

    for checker in checkers:
        checker.join()
    checker_done.set()
    consumer.join()


@tools.command("download-uniprot", short_help=(
    "Downloads a list of proteins from UniProt"))
@click.option("-i", "--name-file-path", help="Read input from a file instead of STDIN")
@click.option("-o", "--output-path", help="Write output to a file instead of STDOUT")
def download_uniprot(name_file_path=None, output_path=None):
    if name_file_path is not None:
        handle = open(name_file_path)
    else:
        handle = sys.stdin

    def name_getter(x):
        header = fasta.default_parser(x)
        try:
            return header.accession
        except Exception:
            return header[0]

    accession_list = []
    for line in handle:
        accession_list.append(name_getter(line))
    if output_path is None:
        outhandle = click.get_binary_stream('stdout')
    else:
        outhandle = open(output_path, 'wb')
    writer = fasta.FastaFileWriter(outhandle)
    downloader = UniprotProteinDownloader(accession_list, 10)
    downloader.start()
    has_work = True

    def make_protein(accession, uniprot_protein):
        sequence = uniprot_protein.sequence
        name = "sp|{accession}|{gene_name} {description}".format(
            accession=accession, gene_name=uniprot_protein.gene_name,
            description=uniprot_protein.recommended_name)
        return name, sequence

    while has_work:
        try:
            accession, value = downloader.get(True, 3)
            if isinstance(value, Exception):
                click.echo("Could not retrieve %s - %r" % (accession, value), err=True)
            else:
                writer.write(*make_protein(accession, value))

        except Empty:
            # If we haven't completed the download process, block
            # now and wait for the threads to join, then continue
            # trying to fetch results
            if not downloader.done_event.is_set():
                downloader.join()
                continue
            # Otherwise we've already waited for all the results to
            # arrive and we can stop iterating
            else:
                has_work = False


@tools.command("mass-search")
@click.option("-p", "--glycopeptide", is_flag=True)
@click.option("-m", "--error-tolerance", type=RelativeMassErrorParam(), default=1e-5)
@click.argument("database-connection")
@click.argument("hypothesis-identifier")
@click.argument("target-mass", type=float)
def mass_search(database_connection, hypothesis_identifier, target_mass, glycopeptide=False, error_tolerance=1e-5):
    if glycopeptide:
        handle = GlycopeptideDiskBackedStructureDatabase(database_connection, int(hypothesis_identifier))
    else:
        handle = GlycanCompositionDiskBackedStructureDatabase(database_connection, int(hypothesis_identifier))
    width = (target_mass * error_tolerance)
    click.secho("Mass Window: %f-%f" % (target_mass - width, target_mass + width), fg='yellow')
    hits = list(handle.search_mass_ppm(target_mass, error_tolerance))
    if not hits:
        click.secho("No Matches", fg='red')
    for hit in hits:
        click.echo("\t".join(map(str, hit)))


@tools.command("version-check")
def version_check():
    packages = [
        "glycresoft",
        "glycresoft_app",
        "glypy",
        "glycopeptidepy",
        "ms_peak_picker",
        "brain-isotopic-distribution",
        "ms_deisotope",
        "pyteomics",
        "lxml",
        "numpy",
        "scipy",
        "matplotlib"
    ]
    click.secho("Library Versions", fg='yellow')
    for dep in packages:
        try:
            rev = pkg_resources.require(dep)
            click.echo(str(rev[0]))
        except Exception:
            try:
                module = __import__(dep)
            except ImportError:
                continue
            version = getattr(module, "__version__", None)
            if version is None:
                version = getattr(module, "version", None)
            if version is None:
                try:
                    module = __import__("%s.version" % dep).version
                    version = module.version
                except ImportError:
                    continue
            if version:
                click.echo("%s %s" % (dep, version))


@tools.command("interactive-shell")
@click.option("-s", "--script", default=None)
@click.argument("script_args", nargs=-1)
def interactive_shell(script_args, script):
    if script:
        with open(script, 'rt') as fh:
            script = fh.read()
        exec(script)
    code.interact(local=locals())


@tools.command("update-analysis-parameters")
@database_connection
@analysis_identifier
@click.option("-p", "--parameter", nargs=2, multiple=True, required=False)
def update_analysis_parameters(database_connection, analysis_identifier, parameter):
    session = database_connection.session
    analysis = get_by_name_or_id(session, Analysis, analysis_identifier)
    click.echo("Current Parameters:")
    pprint.pprint(analysis.parameters)
    for name, value in parameter:
        analysis.parameters[name] = value
    session.add(analysis)
    session.commit()


@tools.command("summarize-glycopeptide-hypothesis",
               short_help="Show summary information about a glycopeptide hypothesis")
@database_connection
@hypothesis_identifier(GlycopeptideHypothesis)
def summarize_glycopeptide_hypothesis(database_connection, hypothesis_identifier):
    session = database_connection.session
    hypothesis = get_by_name_or_id(session, GlycopeptideHypothesis, hypothesis_identifier)
    gp_counts = session.query(Protein, func.count(Glycopeptide.id)).join(
        Glycopeptide).group_by(Protein.id).filter(
        Protein.hypothesis_id == hypothesis.id).order_by(
        Protein.id).all()
    pept_counts = session.query(Protein, func.count(Peptide.id)).join(
        Peptide).group_by(Protein.id).filter(
        Protein.hypothesis_id == hypothesis.id).order_by(
        Protein.id).all()

    gc_count = session.query(GlycanClass.name, func.count(GlycanClass.id)).join(
        GlycanComposition.structure_classes).filter(GlycanComposition.hypothesis_id == hypothesis.id).group_by(
            GlycanClass.id).order_by(GlycanClass.name).all()

    counts = {}
    for prot, c in gp_counts:
        counts[prot.id] = [prot, 0, c]
    for prot, c in pept_counts:
        try:
            counts[prot.id][1] = c
        except KeyError:
            counts[prot.id] = [prot, c, 0]

    counts = sorted(counts.values(), key=lambda x: x[::-1][:-1], reverse=1)
    pept_total = 0
    gp_total = 0
    for protein, pept_count, gp_count in counts:
        click.echo("%s: %d" % (protein.name, pept_count))
        pept_total += pept_count
        gp_total += gp_count

    click.echo("Total Peptides: %d" % (pept_total, ))
    click.echo("Total Glycopeptides: %d" % (gp_total, ))
    for cls_name, count in gc_count:
        click.echo("%s Glycan Compositions: %d" % (cls_name, count))
    click.echo(pprint.pformat(hypothesis.parameters))


@tools.command("extract-blob")
@database_connection
@click.argument("blob-identifier")
@click.option("-o", "--output-path", type=click.File(mode='wb'))
def extract_file_blob(database_connection, blob_identifier, output_path=None):
    if output_path is None:
        output_path = click.get_binary_stream('stdout')
    session = database_connection.session
    blob = get_by_name_or_id(session, FileBlob, blob_identifier)
    with blob.open() as fh:
        chunk_size = 2 ** 16
        chunk = fh.read(chunk_size)
        while chunk:
            output_path.write(chunk)
            chunk = fh.read(chunk_size)


@tools.command("csv-concat")
@click.argument("csv-paths", type=click.File(mode='rt'), nargs=-1)
@click.option("-o", "--output-path", type=click.Path(writable=True), help="Path to write output to")
def csv_concat(csv_paths, output_path=None):
    if output_path is None:
        output_path = '-'
    import csv
    headers = None
    with click.open_file(output_path, mode='wt') as outfh:
        writer = csv.writer(outfh)
        for infh in csv_paths:
            reader = csv.reader(infh)
            _header_line = next(reader)
            if headers is None:
                headers = _header_line
                writer.writerow(headers)
            elif _header_line != headers:
                raise click.ClickException("File %s did not have matching headers (%s != %s)" % (
                    infh.name, _header_line, headers))
            for row in reader:
                writer.writerow(row)
            infh.close()
            outfh.flush()



cdef class CompositionGraphNode(object):
    cdef:
        public object composition
        public ssize_t index
        public str _str
        public EdgeSet edges
        public Py_hash_t _hash
        public double _score
        public double internal_score
        public bint marked

cdef class EdgeSet(object):
    cdef:
        public dict store

    cpdef bint add_if_shorter(self, CompositionGraphEdge edge)
    cdef list get_edges(self)


cdef class CompositionGraphEdge(object):
    cdef:
        public CompositionGraphNode node1
        public CompositionGraphNode node2
        public long order
        public double weight
        public Py_hash_t _hash
        public str _str


    @staticmethod
    cdef CompositionGraphEdge _create(CompositionGraphNode node1, CompositionGraphNode node2, long order, double weight)

    cdef void _init(self)
    cpdef copy_for(self, CompositionGraphNode node1, CompositionGraphNode node2)



cdef class CompositionGraphBase(object):
    cdef:
        public list nodes
        public dict node_map
        public object _composition_normalizer
        public object distance_fn
        public EdgeSet edges
        public object neighborhoods

    cdef list get_edges(self)
    cpdef CompositionGraphBase copy(self)

cpdef reindex_graph(self)


cdef struct glycan_composition_vector:
    int* counts
    size_t size


cdef int initialize_glycan_composition_vector(size_t size, glycan_composition_vector* self) nogil
cdef int destroy_glycan_composition_vector(glycan_composition_vector* self) nogil
cdef int copy_glycan_composition_vector(glycan_composition_vector* self, glycan_composition_vector* into) nogil

cdef double glycan_composition_vector_distance(glycan_composition_vector* self, glycan_composition_vector* other) nogil
cdef int glycan_composition_vector_difference(glycan_composition_vector* self, glycan_composition_vector* other, glycan_composition_vector* into) nogil
cdef int glycan_composition_vector_addition(glycan_composition_vector* self, glycan_composition_vector* other, glycan_composition_vector* into) nogil



cdef class GlycanCompositionVectorContext:
    cdef:
        public list components
        public size_t component_count

    cdef glycan_composition_vector* encode_raw(self, glycan_composition)
    cpdef GlycanCompositionVector encode(self, glycan_composition)
    cpdef decode(self, GlycanCompositionVector gcvec)


cdef class GlycanCompositionVector:
    cdef:
        glycan_composition_vector* ptr

    @staticmethod
    cdef GlycanCompositionVector _create(glycan_composition_vector* ptr)

    cpdef double distance(self, GlycanCompositionVector other) except -1

    cpdef GlycanCompositionVector difference(self, GlycanCompositionVector other)
    cpdef GlycanCompositionVector addition(self, GlycanCompositionVector other)

    cpdef GlycanCompositionVector clone(self)


cdef class CachingGlycanCompositionVectorContext(GlycanCompositionVectorContext):
    cdef:
        public dict encode_cache
        public dict decode_cache


# -*- coding: utf8 -*-
'''This module implements techniques derived from the pGlyco2
FDR estimation procedure described in:

[1] Liu, M.-Q., Zeng, W.-F., Fang, P., Cao, W.-Q., Liu, C., Yan, G.-Q., … Yang, P.-Y.
    (2017). pGlyco 2.0 enables precision N-glycoproteomics with comprehensive quality
    control and one-step mass spectrometry for intact glycopeptide identification.
    Nature Communications, 8(1), 438. https://doi.org/10.1038/s41467-017-00535-2
[2] Zeng, W.-F., Liu, M.-Q., Zhang, Y., Wu, J.-Q., Fang, P., Peng, C., … Yang, P. (2016).
    pGlyco: a pipeline for the identification of intact N-glycopeptides by using HCD-
    and CID-MS/MS and MS3. Scientific Reports, 6(April), 25102. https://doi.org/10.1038/srep25102
'''
from typing import Dict, List, Optional, Tuple, Type, Union
import numpy as np

from glypy.structure.glycan_composition import GlycanComposition
from glypy.utils.enum import EnumValue

try:
    from matplotlib import pyplot as plt
except ImportError:
    pass

from glycresoft.task import TaskBase

from glycresoft.structure.enums import GlycopeptideFDREstimationStrategy
from glycresoft.tandem.spectrum_match.spectrum_match import FDRSet, MultiScoreSpectrumMatch, SpectrumMatch
from glycresoft.tandem.spectrum_match.solution_set import MultiScoreSpectrumSolutionSet, SpectrumSolutionSet
from glycresoft.tandem.target_decoy import (
    NearestValueLookUp,
    PeptideScoreTargetDecoyAnalyzer,
    FDREstimatorBase,
    PeptideScoreSVMModel
)
from glycresoft.tandem.glycopeptide.core_search import approximate_internal_size_of_glycan

from .mixture import (
    GammaMixture, GaussianMixture, GaussianMixtureWithPriorComponent,
    MixtureBase, TruncatedGaussianMixture, TruncatedGaussianMixtureWithPriorComponent)

from .journal import SolutionSetGrouper


def noop(*args, **kwargs):
    pass


def interpolate_from_zero(nearest_value_map: NearestValueLookUp, zero_value: float=1.0) -> NearestValueLookUp:
    smallest = nearest_value_map.items[0]
    X = np.linspace(0, smallest[0])
    Y = np.interp(X, [0, smallest[0]], [zero_value, smallest[1]])
    pairs = list(zip(X, Y))
    pairs.extend(nearest_value_map.items)
    return nearest_value_map.__class__(pairs)


class FiniteMixtureModelFDREstimatorBase(FDREstimatorBase):
    decoy_scores: np.ndarray
    target_scores: np.ndarray
    decoy_mixture: Optional[MixtureBase]
    target_mixture: Optional[MixtureBase]
    fdr_map: Optional[NearestValueLookUp]

    def __init__(self, decoy_scores, target_scores):
        self.decoy_scores = np.array(decoy_scores)
        self.target_scores = np.array(target_scores)
        self.decoy_mixture = None
        self.target_mixture = None
        self.fdr_map = None

    def estimate_posterior_error_probability(self, X: np.ndarray) -> np.ndarray:
        return self.target_mixture.prior.score(X) * self.pi0 / self.target_mixture.score(X)

    def get_count_for_fdr(self, q_value: float):
        target_scores = self.target_scores
        target_scores = np.sort(target_scores)
        fdr = np.sort(self.estimate_fdr())[::-1]
        if len(fdr) == 0:
            return 0, 0
        ii = np.where(fdr < q_value)[0]
        if len(ii) == 0:
            return float('inf'), 0
        i = ii[0]
        score_for_fdr = target_scores[i]
        target_counts = (target_scores >= score_for_fdr).sum()
        return score_for_fdr, target_counts

    @property
    def pi0(self):
        return self.target_mixture.weights[-1]

    def plot_mixture(self, ax=None):
        if ax is None:
            fig, ax = plt.subplots(1)
        X = np.arange(1, max(self.target_scores), 0.1)
        ax.plot(X,
                np.exp(self.target_mixture.logpdf(X)).sum(axis=1))
        for col in np.exp(self.target_mixture.logpdf(X)).T:
            ax.plot(X, col, linestyle='--')
        ax.hist(self.target_scores, bins=100, density=1, alpha=0.15)
        return ax

    def plot(self, ax=None):
        if ax is None:
            fig, ax = plt.subplots(1)

        dmin = 0
        dmax = 1
        tmin = 0
        tmax = 1
        if len(self.target_scores):
            tmin = self.target_scores.min()
            tmax = self.target_scores.max()
        if len(self.decoy_scores):
            dmin = self.decoy_scores.min()
            dmax = self.decoy_scores.max()
        points = np.linspace(
            min(tmin, dmin),
            max(tmax, dmax),
            10000)
        target_scores = np.sort(self.target_scores)
        target_counts = [(self.target_scores >= i).sum() for i in points]
        decoy_counts = [(self.decoy_scores >= i).sum() for i in points]
        fdr = self.estimate_fdr(target_scores)
        at_5_percent = np.where(fdr < 0.05)[0][0]
        at_1_percent = np.where(fdr < 0.01)[0][0]
        line1 = ax.plot(points, target_counts,
                        label='Target', color='steelblue')
        line2 = ax.plot(points, decoy_counts, label='Decoy', color='coral')
        line4 = ax.vlines(target_scores[at_5_percent], 0, np.max(
            target_counts), linestyle='--', color='green', lw=0.75, label='5% FDR')
        line5 = ax.vlines(target_scores[at_1_percent], 0, np.max(
            target_counts), linestyle='--', color='skyblue', lw=0.75, label='1% FDR')
        ax.set_ylabel("# Matches Retained")
        ax.set_xlabel("Score")
        ax2 = ax.twinx()
        ax2.set_ylabel("FDR")
        line3 = ax2.plot(target_scores, fdr, label='FDR',
                         color='grey', linestyle='--')
        ax.legend(
            [line1[0], line2[0], line3[0], line4, line5],
            ['Target', 'Decoy', 'FDR', '5% FDR', '1% FDR'], frameon=False)

        lo, hi = ax.get_ylim()
        lo = max(lo, 0)
        ax.set_ylim(lo, hi)
        lo, hi = ax2.get_ylim()
        ax2.set_ylim(0, hi)

        lo, hi = ax.get_xlim()
        ax.set_xlim(-1, hi)
        lo, hi = ax2.get_xlim()
        ax2.set_xlim(-1, hi)
        return ax

    def fit(self, max_components: int=10,
            max_target_components: Optional[int]=None,
            max_decoy_components: Optional[int]=None) -> NearestValueLookUp:
        if not max_target_components:
            max_target_components = max_components
        if not max_decoy_components:
            max_decoy_components = max_components
        self.estimate_decoy_distributions(max_decoy_components)
        self.estimate_target_distributions(max_target_components)
        fdr = self.estimate_fdr(self.target_scores)
        self.fdr_map = NearestValueLookUp(zip(self.target_scores, fdr))
        # Since 0 is not in the domain of the model, we need to include it by interpolating from 1 to the smallest
        # fitted value.
        if self.fdr_map.items[0][0] > 0:
            self.fdr_map = interpolate_from_zero(self.fdr_map)
        return self.fdr_map

    def estimate_fdr(self, X: np.ndarray=None) -> np.ndarray:
        if X is None:
            X = self.target_scores
        X_ = np.array(sorted(X, reverse=True))
        pep = self.estimate_posterior_error_probability(X_)
        # The FDR is the expected value of PEP, or the average PEP in this case.
        # The expression below is a cumulative mean (the cumulative sum divided
        # by the number of elements in the sum)
        fdr = np.cumsum(pep) / np.arange(1, len(X_) + 1)
        # Use searchsorted on the ascending ordered version of X_
        # to find the indices of the origin values of X, then map
        # those into the ascending ordering of the FDR vector to get
        # the FDR estimates of the original X
        fdr[np.isnan(fdr)] = 1.0
        fdr_descending = fdr[::-1]
        for i in range(1, fdr_descending.shape[0]):
            if fdr_descending[i - 1] < fdr_descending[i]:
                fdr_descending[i] = fdr_descending[i - 1]
        fdr = fdr_descending[::-1]
        fdr = fdr[::-1][np.searchsorted(X_[::-1], X)]
        return fdr

    def _select_best_number_of_components(self, max_components: int,
                                          fitter: Type[MixtureBase],
                                          scores: np.ndarray, **kwargs) -> MixtureBase:
        models = []
        bics = []
        for i in range(1, max_components + 1):
            self.debug("Fitting %d Components" % (i,))
            iteration_kwargs = kwargs.copy()
            iteration_kwargs['n_components'] = i
            model = fitter(scores, **iteration_kwargs)
            bic = model.bic(scores)
            models.append(model)
            bics.append(bic)
            self.debug("BIC: %g" % (bic,))
        i = np.argmin(bics)
        self.debug("Selected %d Components" % (i + 1,))
        return models[i]

    def estimate_decoy_distributions(self, max_components: int=10):
        raise NotImplementedError()

    def estimate_target_distributions(self, max_components: int=10) -> MixtureBase:
        n = len(self.target_scores)
        np.random.seed(n)
        if n < 10:
            self.log("Too few target observations")
            self.target_mixture = GaussianMixtureWithPriorComponent(
                [1.0], [1.0], self.decoy_mixture, [0.5, 0.5])
            return self.target_mixture
        self.target_mixture = self._select_best_number_of_components(
            max_components,
            GaussianMixtureWithPriorComponent.fit,
            self.target_scores,
            prior=self.decoy_mixture,
            deterministic=True)
        return self.target_mixture

    def score(self, spectrum_match: Union[SpectrumMatch, float], assign: bool = False) -> float:
        return self.fdr_map[
            spectrum_match.score
            if isinstance(spectrum_match, SpectrumMatch)
            else spectrum_match]


class FiniteMixtureModelFDREstimatorSeparated(FiniteMixtureModelFDREstimatorBase):
    decoy_model_type = GammaMixture
    target_model_type = GaussianMixture

    def estimate_decoy_distributions(self, max_components: int = 10):
        n = len(self.decoy_scores)
        np.random.seed(n)
        if n < 10:
            self.log("Too few decoy observations")
            self.decoy_mixture = self.decoy_model_type(
                [1.0], [1.0], [1.0])
            return self.decoy_mixture

        self.decoy_mixture = self._select_best_number_of_components(
            max_components, self.decoy_model_type.fit, self.decoy_scores)
        return self.decoy_mixture

    def estimate_target_distributions(self, max_components: int = 10):
        n = len(self.target_scores)
        np.random.seed(n)
        if n < 10:
            self.log("Too few target observations")
            self.target_mixture = self.target_model_type(
                [1.0], [1.0], [1.0])
            return self.target_mixture
        self.target_mixture = self._select_best_number_of_components(
            max_components,
            GaussianMixture.fit,
            self.target_scores,
            deterministic=True)
        return self.target_mixture

    def estimate_posterior_error_probability(self, X: np.ndarray) -> np.ndarray:
        d = self.decoy_mixture.score(X)
        t = self.target_mixture.score(X)
        pi0 = self.pi0
        return (pi0 * d) / (pi0 * d + (1 - pi0) * t)

    def get_count_for_fdr(self, q_value):
        target_scores = self.target_scores
        target_scores = np.sort(target_scores)
        fdr = np.sort(self.estimate_fdr())[::-1]

        i = np.where(fdr < q_value)[0][0]
        score_for_fdr = target_scores[i]
        target_counts = (target_scores >= score_for_fdr).sum()
        return score_for_fdr, target_counts

    @property
    def pi0(self):
        pi0 = len(self.decoy_scores) / len(self.target_scores)
        return pi0


class FiniteMixtureModelFDREstimatorDecoyGamma(FiniteMixtureModelFDREstimatorBase):

    def estimate_decoy_distributions(self, max_components=10):
        n = len(self.decoy_scores)
        np.random.seed(n)
        if n < 10:
            self.log("Too few decoy observations")
            self.decoy_mixture = GammaMixture([1.0], [1.0], [1.0])
            return self.decoy_mixture

        self.decoy_mixture = self._select_best_number_of_components(
            max_components, GammaMixture.fit, self.decoy_scores)
        return self.decoy_mixture


class FiniteMixtureModelFDREstimatorDecoyGaussian(FiniteMixtureModelFDREstimatorBase):

    def estimate_decoy_distributions(self, max_components=10):
        n = len(self.decoy_scores)
        np.random.seed(n)
        if n < 10:
            self.log("Too few decoy observations")
            self.decoy_mixture = GaussianMixture([1.0], [1.0], [1.0])
            return self.decoy_mixture

        self.decoy_mixture = self._select_best_number_of_components(
            max_components, GaussianMixture.fit, self.decoy_scores)
        return self.decoy_mixture


FiniteMixtureModelFDREstimator = FiniteMixtureModelFDREstimatorDecoyGamma


class FiniteMixtureModelFDREstimatorHalfGaussian(FiniteMixtureModelFDREstimatorBase):

    def estimate_decoy_distributions(self, max_components=10):
        n = len(self.decoy_scores)
        np.random.seed(n)
        if n < 10:
            self.log("Too few decoy observations")
            self.decoy_mixture = TruncatedGaussianMixture([0], [1.0], [1.0])
            return self.decoy_mixture

        self.decoy_mixture = self._select_best_number_of_components(
            max_components, TruncatedGaussianMixture.fit, self.decoy_scores)
        return self.decoy_mixture

    def estimate_target_distributions(self, max_components=10):
        n = len(self.target_scores)
        np.random.seed(n)
        if n < 10:
            self.log("Too few target observations")
            self.target_mixture = TruncatedGaussianMixtureWithPriorComponent(
                [0.0], [1.0], self.decoy_mixture, [0.5, 0.5])
            return self.target_mixture
        self.target_mixture = self._select_best_number_of_components(
            max_components,
            TruncatedGaussianMixtureWithPriorComponent.fit,
            self.target_scores,
            prior=self.decoy_mixture,
            deterministic=True)
        return self.target_mixture


class MultivariateMixtureModel(FDREstimatorBase):
    models: List[FiniteMixtureModelFDREstimatorBase]

    def __init__(self, models):
        self.models = models

    def __iter__(self):
        return iter(self.models)

    def __getitem__(self, i):
        return self.models[i]

    def __len__(self):
        return len(self.models)

    def is_combined(self, mixture_model):
        tm = mixture_model.target_mixture
        if hasattr(tm, 'prior'):
            return True
        return False

    def _compute_per_feature_probabilities(self, X: List[np.ndarray] = None) -> Tuple[np.ndarray, np.ndarray]:
        if X is None:
            X = [m.target_scores for m in self.models]

        decoy_series = []
        target_series = []
        for i, (model, x) in enumerate(zip(self.models, X)):
            if self.is_combined(model):
                decoy_p = model.decoy_mixture.score(x)
                w_decoy = model.pi0
                target_decoy_p = model.target_mixture.score(x)
                target_p = (target_decoy_p - w_decoy * decoy_p)
                decoy_p *= w_decoy
            else:
                decoy_p = model.decoy_mixture.score(x)
                target_p = model.target_mixture.score(x)
            decoy_series.append(decoy_p)
            target_series.append(target_p)
        return decoy_series, target_series

    def estimate_pi0(self, X: List[np.ndarray]=None) -> Tuple[float, np.ndarray, np.ndarray]:
        decoy_series, target_series = self._compute_per_feature_probabilities(X)
        decoy = np.prod(decoy_series, axis=0)
        target = np.prod(target_series, axis=0)
        return (decoy / (decoy + target)).mean(), decoy, target

    def _compute_overlap_probabilities(self, decoy_series: List[np.ndarray],
                                       target_series: List[np.ndarray],
                                       pi0s: List[float]) -> List[np.ndarray]:
        mixed_terms = []
        for i, d in enumerate(decoy_series):
            for j, t in enumerate(target_series):
                if i != j:
                    mixed_terms.append(d * t / (1 - pi0s[j]) * pi0s[j])
        return mixed_terms

    def estimate_posterior_error_probability(self, X: List[np.ndarray] = None, correlated: bool=False) -> np.ndarray:
        if X is None:
            X = [m.target_scores for m in self.models]

        decoy_series, target_series = self._compute_per_feature_probabilities(X)
        if correlated:
            # mixed_terms = self._compute_overlap_probabilities(
            #     decoy_series, target_series, [m.pi0 for m in self])
            raise NotImplementedError("Correlated features are not implemented")

        decoy = np.prod(decoy_series, axis=0)
        target = np.prod(target_series, axis=0)
        if self.is_combined(self.models[0]):
            pass
        else:
            pi_0 = self.models[0].pi0
            decoy *= pi_0
            target *= (1 - pi_0)
        if not correlated:
            return decoy / (target + decoy)
        raise NotImplementedError("Correlated features are not implemented")

    def estimate_fdr(self, X: List[np.ndarray] = None, correlated: bool=False):
        if X is None:
            X = [m.target_scores for m in self.models]
        X = np.stack(X, axis=-1)
        X_ = X[np.lexsort(X[:, ::-1].T)[::-1], :]
        pep = self.estimate_posterior_error_probability(list(X_.T))
        fdr = np.cumsum(pep) / np.arange(1, len(X_) + 1)
        fdr[np.isnan(fdr)] = 1.0
        fdr_descending = fdr[::-1]
        for i in range(1, fdr_descending.shape[0]):
            if fdr_descending[i - 1] < fdr_descending[i]:
                fdr_descending[i] = fdr_descending[i - 1]
        fdr = fdr_descending[::-1]
        fdr = fdr[::-1][np.searchsorted(X_[::-1, 0], X[:, 0])]
        return fdr

    def fit(self, correlated: bool = False, *args, **kwargs) -> NearestValueLookUp:
        fdr = self.estimate_fdr(correlated=correlated)
        self.fdr_map = NearestValueLookUp(
            zip(self.models[0].target_scores, fdr))
        # Since 0 is not in the domain of the model, we need to include it by
        # interpolating from 1 to the smallest fitted value.
        if self.fdr_map.items[0][0] > 0:
            self.fdr_map = interpolate_from_zero(self.fdr_map)
        return self.fdr_map

    def get_count_for_fdr(self, q_value):
        target_scores = self.models[0].target_scores
        target_scores = np.sort(target_scores)
        fdr = np.sort(self.estimate_fdr())[::-1]

        i = np.where(fdr < q_value)[0][0]
        score_for_fdr = target_scores[i]
        target_counts = (target_scores >= score_for_fdr).sum()
        return score_for_fdr, target_counts

    def plot(self, ax=None):
        if ax is None:
            fig, ax = plt.subplots(1)
        target_scores = self.models[0].target_scores
        decoy_scores = self.models[0].decoy_scores

        dmin = 0
        dmax = float('inf')
        tmin = 0
        tmax = float('inf')
        if len(target_scores):
            tmin = target_scores.min()
            tmax = target_scores.max()
        if len(decoy_scores):
            dmin = decoy_scores.min()
            dmax = decoy_scores.max()
        points = np.linspace(
            min(tmin, dmin),
            max(tmax, dmax),
            10000)

        target_scores = np.sort(target_scores)
        target_counts = [(target_scores >= i).sum() for i in points]
        decoy_counts = [(decoy_scores >= i).sum() for i in points]

        fdr = np.sort(self.estimate_fdr())[::-1]

        at_5_percent = np.where(fdr < 0.05)[0][0]
        at_1_percent = np.where(fdr < 0.01)[0][0]

        line1 = ax.plot(points, target_counts,
                        label='Target', color='steelblue')
        line2 = ax.plot(points, decoy_counts, label='Decoy', color='coral')
        line4 = ax.vlines(target_scores[at_5_percent], 0, np.max(
            target_counts), linestyle='--', color='green', lw=0.75, label='5% FDR')
        line5 = ax.vlines(target_scores[at_1_percent], 0, np.max(
            target_counts), linestyle='--', color='skyblue', lw=0.75, label='1% FDR')
        ax.set_ylabel("# Matches Retained")
        ax.set_xlabel("Score")
        ax2 = ax.twinx()
        ax2.set_ylabel("FDR")
        line3 = ax2.plot(target_scores, fdr, label='FDR',
                         color='grey', linestyle='--')
        ax.legend(
            [line1[0], line2[0], line3[0], line4, line5],
            ['Target', 'Decoy', 'FDR', '5% FDR', '1% FDR'], frameon=False)

        lo, hi = ax.get_ylim()
        lo = max(lo, 0)
        ax.set_ylim(lo, hi)
        lo, hi = ax2.get_ylim()
        ax2.set_ylim(0, hi)

        lo, hi = ax.get_xlim()
        ax.set_xlim(-1, hi)
        lo, hi = ax2.get_xlim()
        ax2.set_xlim(-1, hi)
        return ax


class GlycanSizeCalculator(object):
    glycan_size_cache: Dict[str, int]

    def __init__(self):
        self.glycan_size_cache = dict()

    def get_internal_size(self, glycan_composition: GlycanComposition) -> int:
        key = str(glycan_composition)
        if key in self.glycan_size_cache:
            return self.glycan_size_cache[key]
        n = approximate_internal_size_of_glycan(glycan_composition)
        self.glycan_size_cache[key] = n
        return n

    def __call__(self, glycan_composition: GlycanComposition) -> int:
        return self.get_internal_size(glycan_composition)


class GlycopeptideFDREstimator(TaskBase):
    display_fields = False
    minimum_glycan_size: int = 3
    minimum_score: float = 1

    strategy: EnumValue
    grouper: 'SolutionSetGrouper'

    glycan_fdr: FiniteMixtureModelFDREstimator
    peptide_fdr: Union[PeptideScoreTargetDecoyAnalyzer, PeptideScoreSVMModel]
    glycopeptide_fdr: FiniteMixtureModelFDREstimator

    _peptide_fdr_estimator_type: Type[PeptideScoreTargetDecoyAnalyzer]

    def __init__(self, groups: 'SolutionSetGrouper',
                 strategy: Optional[GlycopeptideFDREstimationStrategy] = None,
                 peptide_fdr_estimator: Optional[Type[PeptideScoreTargetDecoyAnalyzer]] = None):
        if peptide_fdr_estimator is None:
            peptide_fdr_estimator = PeptideScoreTargetDecoyAnalyzer
        if strategy is None:
            strategy = GlycopeptideFDREstimationStrategy.multipart_gamma_gaussian_mixture
        else:
            strategy = GlycopeptideFDREstimationStrategy[strategy]
        self.strategy = strategy
        self.grouper = groups
        self.glycan_fdr = None
        self.peptide_fdr = None
        self.glycopeptide_fdr = None
        self._peptide_fdr_estimator_type = peptide_fdr_estimator

    def fit_glycan_fdr(self):
        tt_gpsms = self.grouper.exclusive_match_groups['target_peptide_target_glycan']
        td_gpsms = self.grouper.exclusive_match_groups['target_peptide_decoy_glycan']

        target_glycan_scores = np.array(
            [s.score_set.glycan_score for s in tt_gpsms])
        decoy_glycan_scores = np.array(
            [s.score_set.glycan_score for s in td_gpsms])

        sizer = GlycanSizeCalculator()
        size_mask = np.array(
            [sizer(s.target.glycan_composition)
             for s in td_gpsms], dtype=int)

        glycan_fdr = FiniteMixtureModelFDREstimator(
            decoy_glycan_scores[(size_mask > self.minimum_glycan_size) & (
                decoy_glycan_scores > self.minimum_score)],
            target_scores=target_glycan_scores[target_glycan_scores > self.minimum_score])

        glycan_fdr.fit()
        glycan_fdr.summarize("Glycan FDR")

        self.glycan_fdr = glycan_fdr
        return self.glycan_fdr

    def fit_peptide_fdr(self):
        tt_gpsms = self.grouper.exclusive_match_groups['target_peptide_target_glycan']
        dt_gpsms = self.grouper.exclusive_match_groups['decoy_peptide_target_glycan']

        target_peptides = tt_gpsms
        decoy_peptides = dt_gpsms

        peptide_fdr = self._peptide_fdr_estimator_type(target_peptides, decoy_peptides)
        peptide_fdr.summarize("Peptide FDR")

        self.peptide_fdr = peptide_fdr
        return self.peptide_fdr

    def fit_glycopeptide_fdr(self):
        tt_gpsms = self.grouper.exclusive_match_groups['target_peptide_target_glycan']
        dd_gpsms = self.grouper.exclusive_match_groups['decoy_peptide_decoy_glycan']

        target_total_scores = np.array([t.score_set.glycopeptide_score for t in tt_gpsms])
        decoy_total_scores = np.array([t.score_set.glycopeptide_score for t in dd_gpsms])

        glycopeptide_fdr = FiniteMixtureModelFDREstimator(
            decoy_total_scores[decoy_total_scores > self.minimum_score],
            target_total_scores[target_total_scores > self.minimum_score])

        glycopeptide_fdr.fit()
        glycopeptide_fdr.summarize("Glycopeptide FDR")

        self.glycopeptide_fdr = glycopeptide_fdr
        return self.glycopeptide_fdr

    def _assign_joint_fdr(self, solution_sets: List[MultiScoreSpectrumSolutionSet]):
        for ts in solution_sets:
            for t in ts:
                t.q_value_set.peptide_q_value = self.peptide_fdr.score(t, assign=False)
                t.q_value_set.glycan_q_value = self.glycan_fdr.fdr_map[t.score_set.glycan_score]
                t.q_value_set.glycopeptide_q_value = self.glycopeptide_fdr.fdr_map[t.score_set.glycopeptide_score]
                total = (t.q_value_set.peptide_q_value + t.q_value_set.glycan_q_value -
                         t.q_value_set.glycopeptide_q_value)
                total = max(t.q_value_set.peptide_q_value, t.q_value_set.glycan_q_value, total)
                t.q_value = t.q_value_set.total_q_value = total
            if isinstance(ts, SpectrumSolutionSet):
                ts.q_value = ts.best_solution().q_value
        return solution_sets

    def _assign_minimum_fdr(self, solution_sets: List[MultiScoreSpectrumSolutionSet]):
        for ts in solution_sets:
            for t in ts:
                t.q_value_set.peptide_q_value = self.peptide_fdr.score(t, assign=False)
                t.q_value_set.glycan_q_value = self.glycan_fdr.fdr_map[t.score_set.glycan_score]
                t.q_value_set.glycopeptide_q_value = self.glycopeptide_fdr.fdr_map[
                    t.score_set.glycopeptide_score]
                total = min(t.q_value_set.peptide_q_value, t.q_value_set.glycan_q_value)
                t.q_value = t.q_value_set.total_q_value = total
            ts.q_value = ts.best_solution().q_value
        return solution_sets

    def _assign_peptide_fdr(self, solution_sets: List[MultiScoreSpectrumSolutionSet]):
        for ts in solution_sets:
            for t in ts:
                total = t.q_value_set.peptide_q_value = self.peptide_fdr.score(t, assign=False)
                t.q_value_set.glycan_q_value = 0
                t.q_value_set.glycopeptide_q_value = 0
                t.q_value = t.q_value_set.total_q_value = total
            ts.q_value = ts.best_solution().q_value
        return solution_sets

    def _assign_glycan_fdr(self, solution_sets):
        for ts in solution_sets:
            for t in ts:
                t.q_value_set.peptide_q_value = 0
                total = t.q_value_set.glycan_q_value = self.glycan_fdr.fdr_map[t.score_set.glycan_score]
                t.q_value_set.glycopeptide_q_value = 0
                t.q_value = t.q_value_set.total_q_value = total
            ts.q_value = ts.best_solution().q_value
        return solution_sets

    def fit_total_fdr(self, solution_sets: Optional[List[MultiScoreSpectrumSolutionSet]] = None):
        if solution_sets is None:
            solution_sets = self.grouper.match_type_groups['target_peptide_target_glycan']
        if self.strategy == GlycopeptideFDREstimationStrategy.multipart_gamma_gaussian_mixture:
            self._assign_joint_fdr(solution_sets)
        elif self.strategy == GlycopeptideFDREstimationStrategy.peptide_fdr:
            self._assign_peptide_fdr(solution_sets)
        elif self.strategy == GlycopeptideFDREstimationStrategy.glycan_fdr:
            self._assign_glycan_fdr(solution_sets)
        elif self.strategy == GlycopeptideFDREstimationStrategy.peptide_or_glycan:
            self._assign_minimum_fdr(solution_sets)
        else:
            raise NotImplementedError(self.strategy)
        return self.grouper

    def score_all(self, solution_set: MultiScoreSpectrumSolutionSet):
        self.fit_total_fdr([solution_set])
        return solution_set

    def score(self, gpsm: MultiScoreSpectrumMatch, assign: bool=False):
        q_value = gpsm.q_value
        q_value_set = gpsm.q_value_set
        placeholder = FDRSet.default()
        gpsm.q_value_set = placeholder
        self.score_all(MultiScoreSpectrumSolutionSet(gpsm.scan, [gpsm]))
        result_q_value_set = placeholder
        if not assign:
            gpsm.q_value = q_value
            gpsm.q_value_set = q_value_set
        return result_q_value_set

    def run(self):
        if self.strategy in (GlycopeptideFDREstimationStrategy.multipart_gamma_gaussian_mixture,
                             GlycopeptideFDREstimationStrategy.glycan_fdr,
                             GlycopeptideFDREstimationStrategy.peptide_or_glycan):
            self.fit_glycan_fdr()
        if self.strategy in (GlycopeptideFDREstimationStrategy.multipart_gamma_gaussian_mixture,
                             GlycopeptideFDREstimationStrategy.peptide_fdr,
                             GlycopeptideFDREstimationStrategy.peptide_or_glycan):
            self.fit_peptide_fdr()
        if self.strategy in (GlycopeptideFDREstimationStrategy.multipart_gamma_gaussian_mixture,
                             GlycopeptideFDREstimationStrategy.peptide_or_glycan):
            self.fit_glycopeptide_fdr()
        self.fit_total_fdr()
        return self.grouper

    def pack(self):
        self.grouper = SolutionSetGrouper([])
        if self.peptide_fdr is not None:
            self.peptide_fdr.pack()



import hjson
import os
import platform
import shutil
import click


from .peptide_modification import (load_modification_rule, save_modification_rule)
from .substituent import (load_substituent_rule, save_substituent_rule)

from psims.controlled_vocabulary import controlled_vocabulary as cv


CONFIG_DIR = click.get_app_dir("glycresoft")

if platform.system().lower() != 'windows':
    os.environ["NOWAL"] = "1"

_mpl_cache_dir = os.path.join(CONFIG_DIR, 'mpl')
cv_store = os.path.join(CONFIG_DIR, 'cv')


def populate_config_dir():
    os.makedirs(CONFIG_DIR)

    # Pre-populate OBO Cache to avoid needing to look these up
    # by URL later
    os.makedirs(cv_store)
    with open(os.path.join(cv_store, 'psi-ms.obo'), 'wb') as fh:
        fh.write(cv._use_vendored_psims_obo().read())
    with open(os.path.join(cv_store, 'unit.obo'), 'wb') as fh:
        fh.write(cv._use_vendored_unit_obo().read())

    os.makedirs(_mpl_cache_dir)


def delete_config_dir():
    shutil.rmtree(CONFIG_DIR, ignore_errors=True)


if not os.path.exists(CONFIG_DIR):
    populate_config_dir()

if not os.path.exists(_mpl_cache_dir):
    os.makedirs(_mpl_cache_dir)

os.environ["MPLCONFIGDIR"] = _mpl_cache_dir

CONFIG_FILE_NAME = "glycresoft-cfg.hjson"
USER_CONFIG_PATH = os.path.join(CONFIG_DIR, CONFIG_FILE_NAME)

cv.configure_obo_store(os.path.join(CONFIG_DIR, "cv"))

HAS_CONFIG = os.path.exists(USER_CONFIG_PATH)

DEFAULT_CONFIG = {
    "version": 0.4,
    "peptide_modifications": {},
    "glycan_modifications": {},
    "substituent_rules": {},
    "environment": {
        "log_file_name": "glycresoft-log",
        "log_file_mode": "a"
    },
    "xml_huge_tree": True
}

_CURRENT_CONFIG = None

if not HAS_CONFIG:
    hjson.dump(DEFAULT_CONFIG, open(USER_CONFIG_PATH, 'w'))


def process(config):
    for key, value in config['peptide_modifications'].items():
        load_modification_rule(value)

    for key, value in config["substituent_rules"].items():
        load_substituent_rule(value)


def recursive_merge(a, b):
    for k, v in b.items():
        if isinstance(b[k], dict) and isinstance(a.get(k), dict):
            recursive_merge(a[k], v)
        else:
            a[k] = v


def get_configuration():
    global _CURRENT_CONFIG
    _CURRENT_CONFIG = load_configuration_from_path(USER_CONFIG_PATH, apply=False)
    local_config_path = os.path.join(os.getcwd(), CONFIG_FILE_NAME)
    if os.path.exists(local_config_path):
        local_config = load_configuration_from_path(local_config_path, apply=False)
        recursive_merge(_CURRENT_CONFIG, local_config)
    process(_CURRENT_CONFIG)
    return _CURRENT_CONFIG


def load_configuration_from_path(path, apply=True):
    cfg = hjson.load(open(path))
    if apply:
        process(cfg)
    return cfg


def set_configuration(obj):
    global _CURRENT_CONFIG
    _CURRENT_CONFIG = None
    hjson.dump(obj, open(USER_CONFIG_PATH, 'w'))
    return get_configuration()


def add_user_modification_rule(rule):
    serialized = save_modification_rule(rule)
    config = get_configuration()
    config['peptide_modifications'][serialized['full_name']] = serialized
    set_configuration(config)
    return load_modification_rule(serialized)


def add_user_substituent_rule(rule):
    serialized = save_substituent_rule(rule)
    config = get_configuration()
    config['substituent_rules'][serialized['name']] = serialized
    set_configuration(config)
    return load_substituent_rule(serialized)


try:
    get_configuration()
except Exception:
    set_configuration(DEFAULT_CONFIG)


DEBUG_MODE = bool(os.environ.get("GLYCRESOFTDEBUG"))



'''
Provides a simple symbolic algebra engine for expressing relationships between
symbols representing counts in terms of the arithmetic operators addition,
subtraction, multiplication, and division, and conditional operators greater
than, less than equality, as well as compound & and | relationships.
'''
import re

from numbers import Number

from six import string_types as basestring, PY3

from glypy.composition import composition_transform
from glypy.structure import glycan_composition


_FMR = glycan_composition.FrozenMonosaccharideResidue
_MR = glycan_composition.MonosaccharideResidue
_SR = glycan_composition.SubstituentResidue
_MC = glycan_composition.MolecularComposition

_is_numeric = re.compile(r"^\d+(?:\.\d+)?$")

def is_numeric(x):
    return _is_numeric.match(x) is not None

def numerical(x):
    try:
        return int(x)
    except (TypeError, ValueError):
        try:
            return float(x)
        except (TypeError, ValueError):
            pass
    raise ValueError("Cannot coerce {} to numerical type".format(x))


def ensuretext(x):
    if isinstance(x, str):
        return x
    elif PY3:
        try:
            return x.decode("utf8")
        except AttributeError:
            return str(x)
    else:
        return str(x)


class ExpressionBase(object):
    @staticmethod
    def _coerce(v):
        if isinstance(v, ExpressionBase):
            return v
        elif isinstance(v, basestring):
            return SymbolNode.parse(ensuretext(v))
        elif isinstance(v, Number):
            return ValueNode(v)

    def _as_compound(self, op, other):
        return ExpressionNode(self, op, self._coerce(other))

    def __add__(self, other):
        return self._as_compound(Operator.get("+"), other)

    def __sub__(self, other):
        return self._as_compound(Operator.get("-"), other)

    def __mul__(self, other):
        return self._as_compound(Operator.get("*"), other)

    def __div__(self, other):
        return self._as_compound(Operator.get("/"), other)

    def __gt__(self, other):
        return self._as_compound(Operator.get(">"), other)

    def __ge__(self, other):
        return self._as_compound(Operator.get(">="), other)

    def __lt__(self, other):
        return self._as_compound(Operator.get("<"), other)

    def __le__(self, other):
        return self._as_compound(Operator.get("<="), other)

    def __eq__(self, other):
        return self._as_compound(Operator.get("=="), other)

    def __ne__(self, other):
        return self._as_compound(Operator.get("!="), other)

    def __and__(self, other):
        return self._as_compound(Operator.get("and"), other)

    def __or__(self, other):
        return self._as_compound(Operator.get("or"), other)

    def __call__(self, other):
        return self._as_compound(Operator.get("call"), other)

    def get_symbols(self):
        return []

    def evaluate(self, context):
        raise NotImplementedError()

    def itersymbols(self):
        for i in []:
            yield i


class ConstraintExpression(ExpressionBase):
    """
    A wrapper around an ExpressionNode that is callable
    to test for satisfaction.

    Attributes
    ----------
    expression : ExpressionNode
        The symbolic expression that a composition must
        satisfy in order to be retained.
    """
    @classmethod
    def parse(cls, string):
        return ConstraintExpression(ExpressionNode.parse(string))

    @classmethod
    def from_list(cls, sym_list):
        return cls.parse(' '.join(sym_list))

    def __init__(self, expression):
        self.expression = expression

    def __repr__(self):
        return "{}".format(self.expression)

    def evaluate(self, context):
        """
        Test for satisfaction of :attr:`expression`

        Parameters
        ----------
        context : Solution
            Context to evaluate :attr:`expression` in

        Returns
        -------
        bool
        """
        return context[self.expression]

    def __call__(self, context):
        return self.evaluate(context)

    def __and__(self, other):
        """
        Construct an :class:`AndCompoundConstraint` from
        `self` and `other`

        Parameters
        ----------
        other : ConstraintExpression

        Returns
        -------
        AndCompoundConstraint
        """
        return AndCompoundConstraint(self, other)

    def __or__(self, other):
        """
        Construct an :class:`OrCompoundConstraint` from
        `self` and `other`

        Parameters
        ----------
        other : ConstraintExpression

        Returns
        -------
        OrCompoundConstraint
        """
        return OrCompoundConstraint(self, other)

    def __eq__(self, other):
        return self.expression == other.expression

    def __ne__(self, other):
        return self.expression != other.expression

    def get_symbols(self):
        return self.expression.get_symbols()


class SymbolNode(ExpressionBase):
    """
    A node representing a single symbol or variable with a scalar multiplication
    coefficient. When evaluated in the context of a :class:`Solution`, it's value
    will be substituted for the value hashed there.

    A SymbolNode is equal to its symbol string and hashes the same way.

    Attributes
    ----------
    coefficient : int
        Multiplicative coefficient to scale the value by
    symbol : str
        A name
    """
    @classmethod
    def parse(cls, string):
        coef = []
        i = 0
        while i < len(string) and (string[i].isdigit() or string[i] in '-+'):
            coef.append(string[i])
            i += 1
        coef_val = int(''.join(coef) or 1)
        residue_sym = string[i:]
        if residue_sym == "":
            residue_sym = None
        elif string[i] == "(" and string[-1] == ")":
            residue_sym = residue_sym[1:-1]
        return cls(residue_sym, coef_val)

    def __init__(self, symbol, coefficient=1):
        self.symbol = symbol
        self.coefficient = coefficient

    def __hash__(self):
        return hash(self.symbol)

    def __eq__(self, other):
        if isinstance(other, basestring):
            return self.symbol == ensuretext(other)
        else:
            return self.symbol == other.symbol and self.coefficient == other.coefficient

    def __ne__(self, other):
        return not self == other

    def evaluate(self, context):
        return context[self] * self.coefficient

    def __repr__(self):
        if self.symbol is not None:
            if self.coefficient != 1:
                return "{} * ({})".format(self.coefficient, self.symbol)
            else:
                return "({})".format(self.symbol)
        else:
            return "{}".format(self.coefficient)

    def get_symbols(self):
        return [self.symbol]

    def itersymbols(self):
        yield self


class ValueNode(ExpressionBase):
    """
    Represent a single numeric value in an ExpressionNode

    Attributes
    ----------
    value : int
        The value of this node
    """
    coefficient = 1

    @classmethod
    def parse(cls, string):
        try:
            value = int(string.replace(" ", ""))
        except ValueError:
            value = float(string.replace(" ", ""))
        return cls(value)

    def __init__(self, value):
        self.value = value

    def __hash__(self):
        return hash(self.value)

    def __eq__(self, other):
        if isinstance(other, basestring):
            return str(self.value) == ensuretext(other)
        else:
            return self.value == other.value

    def evaluate(self, context):
        return self.value

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        return str(self.value)


def typify(node):
    """Convert the input into the appropriate
    type of ExpressionBase-derived Node. Will
    apply pattern matching to determine if a
    passed string looks like a numerical constant
    or if the input is a numerical constant, return
    the appropriate node type.

    Parameters
    ----------
    node : object
        The object to be converted.

    Returns
    -------
    ExpressionBase
    """
    if isinstance(node, basestring):
        if re.match(r"^(\d+(.\d+)?)$", ensuretext(node).strip()):
            return ValueNode.parse(node)
        else:
            return SymbolNode.parse(node)
    elif isinstance(node, Number):
        return ValueNode(node)
    else:
        return node


def parse_expression(string: str):
    """Parses a str argument into a :class:`ExpressionNode` instance
    containing arbitrarily complex expressions.

    Parameters
    ----------
    string : str
        The text to be parsed

    Returns
    -------
    ExpressionNode

    Raises
    ------
    SyntaxError
        If there are unresolved parentheses leftover after resolving the parse
        stack
    """
    i = 0
    size = len(string)
    paren_level = 0
    current_symbol = ''
    expression_stack = []
    resolver_stack = []
    current_function = ''
    function_stack = []

    while i < size:
        c = string[i]
        if c == " ":
            if current_symbol != "":
                if current_symbol.endswith(",") and current_symbol != ',':
                    expression_stack.append(current_symbol[:-1])
                    expression_stack.append(",")
                else:
                    expression_stack.append(current_symbol)
            current_symbol = ""
        elif c == "(":
            paren_level += 1
            if current_symbol != "":
                if current_function != '':
                    function_stack.append(current_function)
                current_function = current_symbol
            else:
                if current_function != '':
                    function_stack.append(current_function)
                current_function = ''
            current_symbol = ""
            resolver_stack.append(expression_stack)
            expression_stack = []
        elif c == ")":
            paren_level -= 1
            if current_symbol != "":
                if current_symbol.endswith(","):
                    current_symbol = current_symbol[:-1]
                expression_stack.append(current_symbol)
            current_symbol = ""
            term = collapse_expression_sequence(expression_stack)
            expression_stack = resolver_stack.pop()
            if current_function != "":
                if is_numeric(current_function):
                    coef = numerical(current_function)
                    term = EnclosedExpression(term, coef)
                else:
                    fn = SymbolNode(current_function)
                    term = FunctionCallNode(fn, Operator.get("call"), term)
                if function_stack:
                    current_function = function_stack.pop()
                else:
                    current_function = ''
            else:
                term = EnclosedExpression(term)
                if function_stack:
                    current_function = function_stack.pop()
                else:
                    current_function = ''
            expression_stack.append(term)
        else:
            current_symbol += c
        i += 1

    if current_symbol != "":
        expression_stack.append(current_symbol)

    if len(resolver_stack) > 0:
        raise ValueError("Unpaired parenthesis")
    return collapse_expression_sequence(expression_stack)


def is_operator(symbol):
    return symbol in operator_map


def collapse_expression_sequence(expression_sequence):
    stack = []
    i = 0
    size = len(expression_sequence)
    if size == 1:
        return typify(expression_sequence[0])

    while i < size:
        next_term = expression_sequence[i]
        stack.append(next_term)
        if len(stack) == 3:
            node = ExpressionNode(*stack) # pylint: disable=no-value-for-parameter
            if hasattr(node.left, 'op'):
                if node.left.op.precedence < node.op.precedence:
                    node = ExpressionNode(
                        left=node.left.left, op=node.left.op, right=ExpressionNode(
                            left=node.left.right, op=node.op, right=node.right))
            stack = [node]
        i += 1
    if len(stack) != 1:
        raise ValueError("Incomplete Expression: %s" %
                         ' '.join(expression_sequence))
    return stack[0]


class ExpressionNode(ExpressionBase):
    """
    A compound node representing two terms joined by a binary operator

    Attributes
    ----------
    left : ExpressionBase
        The term to the left of the binary operator
    op : Operator or str
        The binary operator being applied. If passed as a str
        the Operator type will be looked up using the registered
        operator map.
    right : ExpressionBase
        The term to the right of the binary operator
    """

    #: Expressions have an implicit constant multiplier of 1
    coefficient = 1

    #: Parses a string-like argument into an ExpressionNode instance
    parse = staticmethod(parse_expression)

    def __init__(self, left, op, right):
        self.left = typify(left)
        try:
            self.op = operator_map[op]
        except KeyError:
            if isinstance(op, Operator):
                self.op = op
            else:
                raise TypeError("Could not determine operator type from %r" % op)
        self.right = typify(right)

    def __repr__(self):
        return "{} {} {}".format(self.left, self.op, self.right)

    def evaluate(self, context):
        """
        Execute :attr:`op` on :attr:`left` and :attr:`right` in `context`

        Parameters
        ----------
        context : Solution

        Returns
        -------
        bool or int
        """
        if not isinstance(context, SymbolContext):
            context = SymbolContext(context)
        try:
            return self.op(self.left, self.right, context)
        except KeyError:
            if isinstance(self.left, ValueNode) and self.left.value == 0:
                return True
            elif isinstance(self.right, ValueNode) and self.right.value == 0:
                return True
            else:
                return False

    def get_symbols(self):
        return self.left.get_symbols() + self.right.get_symbols()

    def itersymbols(self):
        for symbol in self.left.itersymbols():
            yield symbol
        for symbol in self.right.itersymbols():
            yield symbol


class EnclosedExpression(ExpressionBase):
    def __init__(self, expr, coefficient=1):
        self.expr = expr
        self.coefficient = coefficient
        self._simplify()

    def _simplify(self):
        while self.coefficient == 1:
            if isinstance(self.expr, EnclosedExpression):
                inner = self.expr
                self.expr = inner.expr
                self.coefficient = inner.coefficient
                return True
            else:
                break
        return False

    def evaluate(self, context):
        return self.expr.evaluate(context) * self.coefficient

    def get_symbols(self):
        return self.expr.get_symbols()

    def __repr__(self):
        if self.coefficient != 1:
            return "%r(%r)" % (self.coefficient, self.expr)
        return "(%r)" % (self.expr, )

    def itersymbols(self):
        return self.expr.itersymbols()


class FunctionCallNode(ExpressionNode):
    def __repr__(self):
        return "{}({})".format(re.sub(r"\)|\(", '', str(self.left)), self.right)


class SymbolSpace(object):
    """Represent an abstract collection of symbols. Provides a
    notion of definedness for Symbols, without associating an
    explicit value for them.

    Attributes
    ----------
    context : dict
        Internal storage of defined symbols
    """
    def __init__(self, context=None):
        if context is None:
            context = dict()
        self.context = self._format_iterable(context)

    @staticmethod
    def _format_iterable(iterable):
        space = dict()
        for key in iterable:
            if isinstance(key, SymbolNode):
                key = key.symbol
            else:
                key = str(key)
            space[key] = 1
        return space

    def _test_symbols_defined(self, symbols, partial=False):
        if not partial:
            for symbol in symbols:
                if not self.defined(symbol):
                    return False
            return True
        else:
            for symbol in symbols:
                if self.defined(symbol):
                    return True
            return False

    def defined(self, expr, partial=False):
        if isinstance(expr, (list, tuple, set, frozenset)):
            return self._test_symbols_defined(expr, partial)
        if not isinstance(expr, ExpressionBase):
            expr = str(expr)
        if isinstance(expr, basestring):
            expr = parse_expression(expr)
        if isinstance(expr, SymbolNode):
            return expr.symbol in self.context
        if isinstance(expr, ValueNode):
            return True
        if isinstance(expr, (ExpressionNode, EnclosedExpression)):
            symbols = expr.get_symbols()
            return self._test_symbols_defined(symbols, partial)

    def partially_defined(self, expr):
        return self.defined(expr, True)

    def __contains__(self, expr):
        return self.defined(expr)

    def __repr__(self):
        return "{%s}" % ', '.join("%s" % c for c in self.context)

    def __iter__(self):
        return iter(self.context)

    def keys(self):
        return self.context.keys()

    def _ipython_key_completions_(self):
        return list(self.keys())


class SymbolContext(SymbolSpace):
    """
    Expands the notion of :class:`SymbolSpace` to bind values
    to Symbols

    Attributes
    ----------
    context : dict
        Mapping from Symbol to Value
    """

    def __init__(self, context=None):
        if context is None:
            context = dict()
        self.context = self._format_map(context)

    def _normalize(self, symbol):
        try:
            symbol = _FMR.from_iupac_lite(symbol)
            return str(symbol)
        except Exception:
            return symbol

    @staticmethod
    def _format_map(mapping):
        store = dict()
        for key, value in mapping.items():
            if isinstance(key, SymbolNode):
                key = key.symbol
            else:
                key = str(key)
            store[key] = value
        return store

    def __getitem__(self, node):
        """
        Summary

        Parameters
        ----------
        node : ExpressionBase

        Returns
        -------
        name : Number
        """
        if not isinstance(node, ExpressionBase):
            node = str(node)

        if isinstance(node, basestring):
            node = parse_expression(node)

        if isinstance(node, SymbolNode):
            if node.symbol is None:
                return 1
            else:
                try:
                    return self.context[node] * node.coefficient
                except KeyError:
                    normalized = self._normalize(node.symbol)
                    if normalized != node.symbol:
                        if normalized in self.context:
                            return self.context[normalized] * node.coefficient
                    return 0
        elif isinstance(node, ValueNode):
            return node.value
        elif isinstance(node, (ExpressionNode, EnclosedExpression)):
            return node.evaluate(self)
        else:
            raise TypeError("Don't know how to evaluate %r of type %s" % (node, node.__class__))

    def __contains__(self, expr):
        if not isinstance(expr, ExpressionBase):
            expr = str(expr)
        if isinstance(expr, basestring):
            expr = parse_expression(expr)
        if isinstance(expr, SymbolNode):
            return expr.symbol in self.context
        elif isinstance(expr, ValueNode):
            return True
        elif isinstance(expr, (ExpressionNode, EnclosedExpression)):
            return expr.evaluate(self) > 0

    def __repr__(self):
        return "SymbolContext(%r)" % self.context

    def items(self):
        return self.context.items()

    def keys(self):
        return self.context.keys()

    def values(self):
        return self.context.values()


_glycosymbol_map = {}


class GlycanSymbolContext(SymbolContext):
    @staticmethod
    def _format_map(mapping):
        store = dict()
        for key, value in dict(mapping).items():
            if isinstance(key, SymbolNode):
                key = key.symbol
            elif isinstance(key, (str, _MR, _SR)):
                text_key = str(key)
                try:
                    key = _glycosymbol_map[text_key]
                except KeyError:
                    key = _FMR.from_iupac_lite(text_key)
                    is_derivatized = composition_transform.has_derivatization(key)
                    if is_derivatized:
                        key = str(
                            glycan_composition.from_iupac_lite.strip_derivatization(text_key))
                    else:
                        key = text_key
                    _glycosymbol_map[text_key] = key
            elif isinstance(key, _MC):
                key = str(key)
            store[key] = value
        return store

    def serialize(self):
        form = "{%s}" % '; '.join("{}:{}".format(str(k), v) for k, v in sorted(
            self.items(), key=lambda x: _monosaccharide_symbol_orderer(x[0])) if v != 0)
        return form

    @staticmethod
    def normalize_key(composition):
        key_obj = composition.clone()
        key_obj = composition_transform.strip_derivatization(key_obj)
        return str(key_obj)


class _monosaccharide_symbol_orderer_cache(dict):

    def __missing__(self, key):
        value = _FMR.from_iupac_lite(key)
        self[key] = value
        return value

    def __call__(self, key):
        key = str(key)
        return self[key].mass(), key


_monosaccharide_symbol_orderer = _monosaccharide_symbol_orderer_cache()


Solution = SymbolContext

operator_map = {}


def register_operator(cls):
    """
    Summary

    Returns
    -------
    name : TYPE
        Description
    """
    inst = cls()
    operator_map[cls.symbol] = inst
    for alias in getattr(cls, "aliases", []):
        operator_map[alias] = inst
    return cls


@register_operator
class Operator(object):
    """
    A base class for defining binary arithmetic
    operators over expressions. Subclasses should
    provide a :meth:`__call__` method to handle
    a left and right :class:`ExpressionBase` objects
    and a :class:`SymbolContext` instance in which
    to resolve them.

    Attributes
    ----------
    precedence : int
        The operator's precedence, which influences
        the order in which they are applied
    symbol : str
        The textual encoding of the operator
    aliases : list
        A list of alternative symbols for the operator
    """
    symbol = "NoOp"
    precedence = 0

    def __init__(self, symbol=None):
        """
        Summary

        Parameters
        ----------
        symbol : TYPE, optional
            Description
        """
        if symbol is not None:
            self.symbol = symbol

    def __call__(self, left, right, context):
        """
        Summary

        Parameters
        ----------
        left : TYPE
            Description
        right : TYPE
            Description
        context : TYPE
            Description

        Returns
        -------
        name : TYPE
            Description
        """
        return NotImplemented

    def __repr__(self):
        return self.symbol

    def __eq__(self, other):
        """
        Summary

        Parameters
        ----------
        other : TYPE
            Description

        Returns
        -------
        name : TYPE
            Description
        """
        try:
            return self.symbol == other.symbol
        except AttributeError:
            return self.symbol == other

    def __hash__(self):
        return hash(self.symbol)

    operator_map = operator_map

    @classmethod
    def get(cls, symbol):
        return cls.operator_map[symbol]


@register_operator
class LessThan(Operator):
    symbol = "<"

    def __call__(self, left, right, context):
        left_val = context[left] * left.coefficient
        right_val = context[right] * right.coefficient
        return left_val < right_val


@register_operator
class LessThanOrEqual(Operator):
    symbol = "<="

    def __call__(self, left, right, context):
        left_val = context[left] * left.coefficient
        right_val = context[right] * right.coefficient
        return left_val <= right_val


@register_operator
class GreaterThan(Operator):
    symbol = ">"

    def __call__(self, left, right, context):
        left_val = context[left] * left.coefficient
        right_val = context[right] * right.coefficient
        return left_val > right_val


@register_operator
class GreaterThanOrEqual(Operator):
    symbol = ">="

    def __call__(self, left, right, context):
        left_val = context[left] * left.coefficient
        right_val = context[right] * right.coefficient
        return left_val >= right_val


@register_operator
class Equal(Operator):
    symbol = "="
    precedence = 10
    aliases = ["=="]

    def __call__(self, left, right, context):
        left_val = context[left] * left.coefficient
        right_val = context[right] * right.coefficient
        return left_val == right_val


@register_operator
class Unequal(Operator):
    symbol = "!="
    precedence = 10

    def __call__(self, left, right, context):
        left_val = context[left] * left.coefficient
        right_val = context[right] * right.coefficient
        return left_val != right_val


@register_operator
class Subtraction(Operator):
    """
    Summary

    Attributes
    ----------
    precedence : int
        Description
    symbol : str
        Description
    """
    symbol = "-"
    precedence = 1

    def __call__(self, left, right, context):
        left_val = context[left]
        right_val = context[right]
        return left_val - right_val


@register_operator
class Addition(Operator):
    symbol = "+"
    precedence = 1

    def __call__(self, left, right, context):
        left_val = context[left]
        right_val = context[right]
        return left_val + right_val


@register_operator
class Multplication(Operator):
    symbol = '*'
    precedence = 2

    def __call__(self, left, right, context):
        left_val = context[left]
        right_val = context[right]
        return left_val * right_val


@register_operator
class Division(Operator):
    symbol = '/'
    precedence = 2

    def __call__(self, left, right, context):
        left_val = context[left]
        right_val = context[right]
        return left_val / right_val


@register_operator
class Or(Operator):
    symbol = 'or'
    aliases = ['|', "||"]

    def __call__(self, left, right, context):
        return context[left] or context[right]


@register_operator
class And(Operator):
    symbol = "and"
    aliases = ["&", "&&"]

    def __call__(self, left, right, context):
        return context[left] and context[right]


@register_operator
class Append(Operator):
    symbol = ','

    def __call__(self, left, right, context):
        lval = context[left]
        rval = context[right]
        if not isinstance(lval, tuple):
            lval = (lval,)
        if not isinstance(rval, tuple):
            rval = (rval,)
        return lval + rval


@register_operator
class Call(Operator):
    symbol = "call"

    def __call__(self, left, right, context):
        try:
            fn = function_table[left]
            return fn(right, context)
        except KeyError:
            raise UnknownFunctionError(left)


class UnknownFunctionError(KeyError):
    pass


def symbolic_sum(terms, context):
    value = 0
    try:
        for term in context[terms]:
            value += context[term]
        return value
    except TypeError:
        return context[terms]


def symbolic_abs(terms, context):
    return abs(context[terms])


def symbolic_max(terms, context):
    args = terms.evaluate(context)
    return max(args)


def symbolic_min(terms, context):
    args = terms.evaluate(context)
    return min(args)


def ifelse(terms, context):
    if len(terms) != 3:
        raise ValueError("ifelse takes three arguments, given %d" % (len(terms), ))
    condition = terms[0]
    if_true = terms[1]
    if_false = terms[2]
    if condition.evaluate(context):
        return if_true.evaluate(context)
    else:
        return if_false.evaluate(context)


function_table = {
    "sum": symbolic_sum,
    "abs": symbolic_abs,
    "max": symbolic_max,
    "min": symbolic_min,
    "ifelse": ifelse
}


class OrCompoundConstraint(ConstraintExpression):
    """
    A combination of ConstraintExpression instances which
    is satisfied if either of its components are satisfied.

    Attributes
    ----------
    left : ConstraintExpression
    right : ConstraintExpression
    """

    def __init__(self, left, right):
        self.left = left
        self.right = right

    def __call__(self, context):
        return self.left(context) or self.right(context)

    def __repr__(self):
        return "(({}) or ({}))".format(self.left, self.right)

    def get_symbols(self):
        return self.left.get_symbols() + self.right.get_symbols()


class AndCompoundConstraint(ConstraintExpression):
    """
    A combination of ConstraintExpression instances which
    is only satisfied if both of its components are satisfied.

    Attributes
    ----------
    left : ConstraintExpression
    right : ConstraintExpression
    """

    def __init__(self, left=None, right=None):
        self.left = left
        self.right = right

    def __call__(self, context):
        return self.left(context) and self.right(context)

    def __repr__(self):
        return "(({}) and ({}))".format(self.left, self.right)

    def get_symbols(self):
        return self.left.get_symbols() + self.right.get_symbols()



from collections import defaultdict
from typing import Any, DefaultDict, Dict, FrozenSet, Iterable, List, Tuple, Union, MutableSequence
from statistics import median

from dataclasses import dataclass, field

from glycopeptidepy import PeptideSequence
from glycopeptidepy.structure.fragment import SimpleFragment

from glypy.structure.glycan_composition import FrozenMonosaccharideResidue, Composition

from glycresoft.tandem.glycopeptide.core_search import GlycanCombinationRecord

from ms_deisotope.peak_set import DeconvolutedPeakSet


class _mass_wrapper(object):

    def __init__(self, mass, annotation=None):
        self.value = mass
        self.annotation = annotation if annotation is not None else mass

    def mass(self, *args, **kwargs):
        return self.value

    def __repr__(self):
        return "MassWrapper(%f, %s)" % (self.value, self.annotation)


_hexnac = FrozenMonosaccharideResidue.from_iupac_lite("HexNAc")
_hexose = FrozenMonosaccharideResidue.from_iupac_lite("Hex")
_neuac = FrozenMonosaccharideResidue.from_iupac_lite('NeuAc')
_neugc = FrozenMonosaccharideResidue.from_iupac_lite("NeuGc")
_water = Composition("H2O").mass

_standard_oxonium_ions = [
    _hexnac,
    _hexose,
    _mass_wrapper(_hexose.mass() - _water),
    _neuac,
    _mass_wrapper(_neuac.mass() - _water),
    FrozenMonosaccharideResidue.from_iupac_lite("Fuc"),
    _mass_wrapper(_hexnac.mass() - Composition("C2H6O3").mass),
    _mass_wrapper(_hexnac.mass() - Composition("CH6O3").mass),
    _mass_wrapper(_hexnac.mass() - Composition("C2H4O2").mass),
    _mass_wrapper(_hexnac.mass() - _water),
    _mass_wrapper(_hexnac.mass() - Composition("H4O2").mass)
]

standard_oxonium_ions = _standard_oxonium_ions[:]

_gscore_oxonium_ions = [
    _hexnac,
    _mass_wrapper(_hexnac.mass() - Composition("C2H6O3").mass),
    _mass_wrapper(_hexnac.mass() - Composition("CH6O3").mass),
    _mass_wrapper(_hexnac.mass() - Composition("C2H4O2").mass),
    _mass_wrapper(_hexnac.mass() - _water),
    _mass_wrapper(_hexnac.mass() - Composition("H4O2").mass)
]


class OxoniumIonScanner(object):

    def __init__(self, ions_to_search=None):
        if ions_to_search is None:
            ions_to_search = _standard_oxonium_ions
        self.ions_to_search = ions_to_search

    def scan(self, peak_list, charge=0, error_tolerance=2e-5, minimum_mass=0):
        matches = []
        for ion in self.ions_to_search:
            if ion.mass() < minimum_mass:
                continue
            match = peak_list.has_peak(
                ion.mass(charge=charge), error_tolerance)
            if match is not None:
                matches.append(match)
        return matches

    def ratio(self, peak_list, charge=0, error_tolerance=2e-5, minimum_mass=0):
        try:
            maximum = max(p.intensity for p in peak_list)
        except ValueError:
            return 0
        n = len([i for i in self.ions_to_search if i.mass() > minimum_mass])
        if n == 0:
            return 0
        oxonium = sum(
            p.intensity / maximum for p in self.scan(
                peak_list, charge, error_tolerance, minimum_mass))
        return oxonium / n

    def __call__(self, peak_list, charge=0, error_tolerance=2e-5, minimum_mass=0):
        return self.ratio(peak_list, charge, error_tolerance, minimum_mass)


oxonium_detector = OxoniumIonScanner()
gscore_scanner = OxoniumIonScanner(_gscore_oxonium_ions)


class SignatureSpecification(object):
    __slots__ = ('components', 'masses', '_hash', '_is_compound')

    def __init__(self, components, masses):
        self.components = tuple(
            FrozenMonosaccharideResidue.from_iupac_lite(k) for k in components)
        self.masses = tuple(masses)
        self._hash = hash(self.components)
        self._is_compound = len(self.masses) > 1

    def __getitem__(self, i):
        return self.components[i]

    def __iter__(self):
        return iter(self.components)

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        return self.components == other.components

    def __repr__(self):
        return "{self.__class__.__name__}({self.components}, {self.masses})".format(self=self)

    def is_expected(self, glycan_composition):
        is_expected = glycan_composition._getitem_fast(self.components[0]) != 0
        if is_expected and self._is_compound:
            is_expected = all(glycan_composition._getitem_fast(
                k) != 0 for k in self.components)
        return is_expected

    def count_of(self, glycan_composition):
        limit = float('inf')
        for component in self:
            cnt = glycan_composition._getitem_fast(component)
            if cnt < limit:
                limit = cnt
        return limit

    def peak_of(self, spectrum, error_tolerance):
        best_peak = None
        best_signal = -1
        for mass in self.masses:
            peaks = spectrum.all_peaks_for(mass, error_tolerance)
            for peak in peaks:
                if peak.intensity > best_signal:
                    best_peak = peak
                    best_signal = peak.intensity
        return best_peak


try:
    _has_c = True
    from glycresoft._c.tandem.oxonium_ions import SignatureSpecification
except ImportError:
    _has_c = False


single_signatures = {
    SignatureSpecification((str(_neuac), ), [
        _neuac.mass(),
        _neuac.mass() - _water
    ]): 0.5,
    SignatureSpecification((str(_neugc), ), [
        _neugc.mass(),
        _neugc.mass() - _water
    ]): 0.5,
}


compound_signatures = {
    SignatureSpecification(('@phosphate', 'Hex'), [
        242.01915393925,
        224.00858925555,
    ]): 0.5,
    SignatureSpecification(("@acetyl", "NeuAc"), [
        333.10598119017,
        315.09541650647
    ]): 0.5
}


class OxoniumIndex(object):
    '''An index for quickly matching all oxonium ions against a spectrum and efficiently mapping them
    back to individual glycan compositions.
    '''
    fragments: List[SimpleFragment]
    fragment_index: DefaultDict[str, List[int]]
    glycan_to_index: Dict[str, int]
    index_to_glycan: Dict[int, Union[str, List[str]]]

    def __init__(self, fragments=None, fragment_index=None, glycan_to_index=None):
        self.fragments = fragments or []
        self.fragment_index = defaultdict(list, fragment_index or {})
        self.glycan_to_index = glycan_to_index or {}
        self.index_to_glycan = {v: k for k, v in self.glycan_to_index.items()}

    def _make_glycopeptide_stub(self, glycan_composition) -> PeptideSequence:
        p = PeptideSequence("P%s" % glycan_composition)
        return p

    def build_index(self, glycan_composition_records: List[GlycanCombinationRecord], **kwargs):
        fragments = {}
        fragment_index = defaultdict(list)
        glycan_index = {}
        for gc_rec in glycan_composition_records:
            glycan_index[gc_rec.composition] = gc_rec.id

            p = self._make_glycopeptide_stub(gc_rec.composition)
            for frag in p.glycan_fragments(**kwargs):
                fragments[frag.name] = frag
                fragment_index[frag.name].append(gc_rec.id)

        self.glycan_to_index = glycan_index
        self.fragment_index = fragment_index
        self.fragments = sorted(fragments.values(), key=lambda x: x.mass)
        self.index_to_glycan = {v: k for k, v in self.glycan_to_index.items()}
        self.simplify()

    def match(self, spectrum: DeconvolutedPeakSet, error_tolerance: float) -> DefaultDict[int, List[Tuple[SimpleFragment, float]]]:
        match_index = defaultdict(list)
        for fragment in self.fragments:
            peak = spectrum.has_peak(fragment.mass, error_tolerance)
            if peak is not None:
                for key in self.fragment_index[fragment.name]:
                    match_index[key].append((fragment, peak.index.neutral_mass))
        return match_index

    def simplify(self):
        id_to_frag_group = defaultdict(set)
        for f, members in self.fragment_index.items():
            for member in members:
                id_to_frag_group[member].add(f)

        groups = defaultdict(list)
        for member, group in id_to_frag_group.items():
            groups[frozenset(group)].append(member)

        counter = 0
        new_fragment_index = defaultdict(list)
        new_glycan_index = {}
        for frag_group, members in groups.items():
            new_id = counter
            counter += 1
            for frag in frag_group:
                new_fragment_index[frag].append(new_id)
            for member in members:
                new_glycan_index[self.index_to_glycan[member]] = new_id

        self.glycan_to_index = new_glycan_index
        self.fragment_index = new_fragment_index
        self.index_to_glycan = defaultdict(list)
        for k, v in self.glycan_to_index.items():
            self.index_to_glycan[v].append(k)


try:
    from glycresoft._c.tandem.oxonium_ions import OxoniumIndex, SignatureIonIndex, SignatureIonIndexMatch
except ImportError:
    SignatureIonIndex = object
    SignatureIonIndexMatch = object


@dataclass(frozen=True)
class OxoniumFilterState:
    scan_id: str
    g_score: float = field(hash=False)
    g_score_pass: bool
    oxonium_ions: FrozenSet[str]


@dataclass
class OxoniumFilterReport(MutableSequence[OxoniumFilterState]):
    records: List[OxoniumFilterState] = field(default_factory=list)

    def copy(self):
        return self.__class__(self.records.copy())

    def __getitem__(self, i):
        return self.records[i]

    def __setitem__(self, i, v):
        self.records[i] = v

    def __delitem__(self, i):
        del self.records[i]

    def append(self, value: OxoniumFilterState):
        self.records.append(value)

    def extend(self, values: Iterable[OxoniumFilterState]):
        self.records.extend(values)

    def insert(self, index: int, value: OxoniumFilterState):
        self.records.insert(index, value)

    def __len__(self):
        return len(self.records)

    def __iter__(self):
        return iter(self.records)

    def prepare_report(self):
        passing: List[OxoniumFilterState] = []
        failing = []
        passing_scores=  []
        for rec in self:
            if rec.g_score_pass:
                passing.append(rec)
                passing_scores.append(rec.g_score)
            else:
                failing.append(rec)
        {
            "n_passing": len(passing),
            "n_failing": len(failing),
            "passing_distribution": passing_scores,
            "passing_median_score": median(passing_scores) if passing_scores else None
        }



'''A set of base implementations for summarizing identified structures
with both MS1 and MSn.
'''
from typing import Any, List, Optional, Set
import numpy as np

from glycresoft.chromatogram_tree import get_chromatogram, ArithmeticMapping, ChromatogramInterface
from glycresoft.serialize.chromatogram import ChromatogramSolution
from glycresoft.tandem.spectrum_match.solution_set import SpectrumSolutionSet
from glycresoft.tandem.spectrum_match.spectrum_match import FDRSet, ScoreSet, SpectrumMatch

from .chromatogram_mapping import TandemSolutionsWithoutChromatogram, TargetType


class IdentifiedStructure(object):
    """A base class for summarizing structures identified by tandem MS
    and matched to MS1 signal.

    Attributes
    ----------
    structure: object
        The structure object identified.
    spectrum_matches: :class:`list` of :class:`~.SolutionSet`
        The MSn spectra that were matched. This is a list of solution sets,
        so other structures may be present, and in the case of multi-score
        matches, they may even appear to score higher.
    chromatogram: :class:`~.ChromatogramSolution`
        The MS1 signal over time matched.
    ms2_score: float
        The best MSn identification score. Usually the highest score amongst
        all spectra matched.
    q_value: float
        The best FDR estimate for this structure. Always the lowest amongst
        all spectra matched.
    ms1_score: float
        The score of :attr:`chromatogram`, or :const:`0` if it is :const:`None`.
    total_signal: float
        The total abundance of :attr:`chromatogram`.
    shared_with: list
        Other :class:`IdentifiedStructure` instances which match the same signal.
    """

    structure: TargetType

    chromatogram: ChromatogramSolution

    spectrum_matches: List[SpectrumSolutionSet]
    _best_spectrum_match: Optional[SpectrumMatch]

    ms1_score: float
    ms2_score: float
    q_value: float

    charge_states: Set[int]
    shared_with: List

    def __init__(self, structure, spectrum_matches, chromatogram, shared_with=None):
        if shared_with is None:
            shared_with = []
        self.structure = structure
        self.spectrum_matches = spectrum_matches
        self.chromatogram = chromatogram
        self._best_spectrum_match = None
        self.ms2_score = None
        self.q_value = None
        self._find_best_spectrum_match()
        self.ms1_score = chromatogram.score if chromatogram is not None else 0
        self.charge_states = chromatogram.charge_states if chromatogram is not None else {
            psm.scan.precursor_information.charge for psm in spectrum_matches
        }
        self.shared_with = shared_with

    def clone(self):
        dup = self.__class__(
            self.structure,
            [sset.clone() for sset in self.spectrum_matches],
            self.chromatogram.clone(),
            self.shared_with[:])
        return dup

    def is_multiscore(self):
        """Check whether this match has been produced by summarizing a multi-score
        match, rather than a single score match.

        Returns
        -------
        bool
        """
        try:
            return self.spectrum_matches[0].is_multiscore()
        except IndexError:
            return False

    @property
    def score_set(self) -> ScoreSet:
        """The :class:`~.ScoreSet` of the best MS/MS match

        Returns
        -------
        :class:`~.ScoreSet`
        """
        if not self.is_multiscore():
            return None
        best_match = self._best_spectrum_match
        if best_match is None:
            return None
        return best_match.score_set

    @property
    def best_spectrum_match(self):
        return self._best_spectrum_match

    @property
    def q_value_set(self) -> FDRSet:
        """The :class:`~.FDRSet` of the best MS/MS match

        Returns
        -------
        :class:`~.FDRSet`
        """
        if not self.is_multiscore():
            return None
        best_match = self._best_spectrum_match
        if best_match is None:
            return None
        return best_match.q_value_set

    @property
    def key(self):
        return self.chromatogram.key

    def _find_best_spectrum_match(self):
        is_multiscore = self.is_multiscore()
        best_match = None
        if is_multiscore:
            best_score = -float('inf')
            best_q_value = float('inf')
        else:
            best_score = -float('inf')
        for solution_set in self.spectrum_matches:
            try:
                match = solution_set.solution_for(self.structure)
                if is_multiscore:
                    q_value = match.q_value
                    if q_value <= best_q_value:
                        q_delta = abs(best_q_value - q_value)
                        best_q_value = q_value
                        if q_delta > 0.001:
                            best_score = match.score
                            best_match = match
                        else:
                            score = match.score
                            if score > best_score:
                                best_score = score
                                best_match = match
                else:
                    score = match.score
                    if score > best_score:
                        best_score = score
                        best_match = match
            except KeyError:
                continue
        if best_match is None:
            self._best_spectrum_match = None
            self.ms2_score = 0
            self.q_value = 1.0
        else:
            self._best_spectrum_match = best_match
            self.ms2_score = best_match.score
            self.q_value = best_match.q_value
        return best_match

    @property
    def neutral_mass(self):
        return self.observed_neutral_mass

    @property
    def observed_neutral_mass(self):
        try:
            return self.chromatogram.neutral_mass
        except AttributeError:
            return self.spectrum_matches[0].scan.precursor_information.neutral_mass

    @property
    def weighted_neutral_mass(self):
        try:
            return self.chromatogram.weighted_neutral_mass
        except AttributeError:
            return self.observed_neutral_mass

    @property
    def start_time(self) -> float:
        try:
            return self.chromatogram.start_time
        except AttributeError:
            return None

    @property
    def end_time(self) -> float:
        try:
            return self.chromatogram.end_time
        except AttributeError:
            return None

    @property
    def apex_time(self) -> float:
        try:
            return self.chromatogram.apex_time
        except AttributeError:
            return None

    def as_arrays(self):
        try:
            return self.chromatogram.as_arrays()
        except AttributeError:
            return np.array([]), np.array([])

    @property
    def integrated_abundance(self):
        if self.chromatogram is None:
            return 0
        return self.chromatogram.integrated_abundance

    def integrate(self):
        if self.chromatogram is None:
            return 0
        return self.chromatogram.integrate()

    @property
    def total_signal(self):
        if self.chromatogram is None:
            return 0
        return self.chromatogram.total_signal

    @property
    def mass_shifts(self):
        try:
            return self.chromatogram.mass_shifts
        except AttributeError:
            shifts = set()
            for solution in self.tandem_solutions:
                try:
                    sm = solution.solution_for(self.structure)
                    shifts.add(sm.mass_shift)
                except KeyError:
                    continue
            return list(shifts)

    def mass_shift_signal_fractions(self):
        try:
            return self.chromatogram.mass_shift_signal_fractions()
        except AttributeError:
            return ArithmeticMapping()

    def __repr__(self):
        return "IdentifiedStructure(%s, %0.3f, %0.3f, %0.3e)" % (
            self.structure, self.ms2_score, self.ms1_score, self.total_signal)

    def get_chromatogram(self):
        return self.chromatogram

    def is_distinct(self, other):
        return get_chromatogram(self).is_distinct(get_chromatogram(other))

    @property
    def tandem_solutions(self):
        return self.spectrum_matches

    @tandem_solutions.setter
    def tandem_solutions(self, value):
        self.spectrum_matches = value

    def mass_shift_tandem_solutions(self):
        mapping = ArithmeticMapping()
        for sm in self.tandem_solutions:
            try:
                mapping[sm.solution_for(self.structure).mass_shift] += 1
            except KeyError:
                continue
        return mapping

    @property
    def glycan_composition(self):
        return self.structure.glycan_composition

    def __eq__(self, other):
        try:
            structure_eq = self.structure == other.structure
            if structure_eq:
                chromatogram_eq = self.chromatogram == other.chromatogram
                if chromatogram_eq:
                    spectrum_matches_eq = self.spectrum_matches == other.spectrum_matches
                    return spectrum_matches_eq
            return False
        except AttributeError:
            return False

    def __hash__(self):
        return hash((self.structure, self.chromatogram))

    def __iter__(self):
        if self.chromatogram is None:
            return iter(())
        return iter(self.chromatogram)

    def __len__(self):
        if self.chromatogram is None:
            return 0
        return len(self.chromatogram)

    def __getitem__(self, i):
        if self.chromatogram is None:
            raise IndexError(i)
        return self.chromatogram[i]

    def overlaps_in_time(self, x):
        if self.chromatogram is None:
            return False
        return self.chromatogram.overlaps_in_time(x)

    @property
    def composition(self):
        return self.chromatogram.composition

    def has_scan(self, scan_id: str) -> bool:
        return any([sset.scan_id == scan_id for sset in self.tandem_solutions])

    def get_scan(self, scan_id: str) -> SpectrumMatch:
        for sset in self.tandem_solutions:
            if sset.scan_id == scan_id:
                return sset
        raise KeyError(scan_id)

    def has_chromatogram(self):
        return self.chromatogram is not None

ChromatogramInterface.register(IdentifiedStructure)


def extract_identified_structures(tandem_annotated_chromatograms, threshold_fn, result_type=IdentifiedStructure):
    identified_structures = []
    unassigned = []

    for chroma in tandem_annotated_chromatograms:
        if chroma.composition is not None:
            if hasattr(chroma, 'entity'):
                try:
                    representers = chroma.representative_solutions
                    if representers is None:
                        representers = chroma.most_representative_solutions(threshold_fn)
                except AttributeError:
                    representers = chroma.most_representative_solutions(threshold_fn)
                bunch = []
                if isinstance(chroma, TandemSolutionsWithoutChromatogram):
                    chromatogram_entry = None
                else:
                    chromatogram_entry = chroma
                for representer in representers:
                    ident = result_type(representer.solution, chroma.tandem_solutions,
                                        chromatogram_entry, [])
                    bunch.append(ident)
                for i, ident in enumerate(bunch):
                    ident.shared_with = bunch[:i] + bunch[i + 1:]
                identified_structures.extend(bunch)
            else:
                unassigned.append(chroma)
        else:
            unassigned.append(chroma)
    return identified_structures, unassigned



import re
import io
import warnings

from typing import Dict, List, Optional, Mapping

from glycopeptidepy.structure.parser import strip_modifications
from glypy.structure.glycan_composition import HashableGlycanComposition

from glycresoft.structure import LRUMapping

from glycresoft.structure.structure_loader import FragmentCachingGlycopeptide
from glycresoft.tandem.spectrum_match import SpectrumMatchClassification as StructureClassification

from .glycosite_model import MINIMUM, GlycosylationSiteModel
from .glycoprotein_model import (
    GlycoproteinSiteSpecificGlycomeModel,
    ReversedProteinSiteReflectionGlycoproteinSiteSpecificGlycomeModel
)


class GlycoproteomeModelBase(object):
    __slots__ = ()
    def score(self, glycopeptide: FragmentCachingGlycopeptide,
              glycan_composition: Optional[HashableGlycanComposition]=None) -> float:
        raise NotImplementedError()

    @classmethod
    def bind_to_hypothesis(cls, session, site_models, hypothesis_id=1, fuzzy=True,
                           site_model_cls=GlycoproteinSiteSpecificGlycomeModel) -> 'GlycoproteomeModelBase':
        inst = cls(
            site_model_cls.bind_to_hypothesis(
                session, site_models, hypothesis_id, fuzzy))
        return inst


class GlycoproteomeModel(GlycoproteomeModelBase):
    glycoprotein_models: Dict[int, GlycoproteinSiteSpecificGlycomeModel]

    def __init__(self, glycoprotein_models):
        if isinstance(glycoprotein_models, Mapping):
            self.glycoprotein_models = dict(glycoprotein_models)
        else:
            self.glycoprotein_models = {
                ggm.id: ggm for ggm in glycoprotein_models
            }

    def relabel_proteins(self, name_to_id_map: Dict[str, int]):
        remapped = {}
        for ggm in self.glycoprotein_models.values():
            try:
                new_id = name_to_id_map[ggm.name]
                remapped[new_id] = ggm
            except KeyError:
                warnings.warn(
                    "No mapping for %r, it will be omitted" % (ggm.name, ))
        self.glycoprotein_models = remapped

    def stub_proteins(self):
        for model in self.glycoprotein_models.values():
            model.stub_protein()

    def find_model(self, glycopeptide: FragmentCachingGlycopeptide):
        if glycopeptide.protein_relation is None:
            return None
        protein_id = glycopeptide.protein_relation.protein_id
        glycoprotein_model = self.glycoprotein_models.get(protein_id)
        return glycoprotein_model

    def score(self, glycopeptide: FragmentCachingGlycopeptide,
              glycan_composition: Optional[HashableGlycanComposition]=None) -> float:
        glycoprotein_model = self.find_model(glycopeptide)
        if glycoprotein_model is None:
            score = MINIMUM
        else:
            score = glycoprotein_model.score(glycopeptide, glycan_composition)
        return score


class SubstringGlycoproteomeModel(GlycoproteomeModelBase):
    sequence_to_model: Dict[str, GlycoproteinSiteSpecificGlycomeModel]

    def __init__(self, models, cache_size=2**15):
        self.models = models
        self.sequence_to_model = {
            str(model.protein): model for model in models.values()
        }
        self.peptide_to_protein = LRUMapping(cache_size)

    def get_models(self, glycopeptide: FragmentCachingGlycopeptide) -> List[GlycoproteinSiteSpecificGlycomeModel]:
        out = []
        seq = strip_modifications(glycopeptide)
        if seq in self.peptide_to_protein:
            return list(self.peptide_to_protein[seq])
        pattern = re.compile(seq)
        for case in self.sequence_to_model:
            if seq in case:
                bounds = pattern.finditer(case)
                for match in bounds:
                    protein_model = self.sequence_to_model[case]
                    site_models = protein_model.find_sites_in(
                        match.start(), match.end())
                    out.append(site_models)
        self.peptide_to_protein[seq] = tuple(out)
        return out

    def find_proteins(self, glycopeptide: FragmentCachingGlycopeptide) -> List[GlycoproteinSiteSpecificGlycomeModel]:
        out = []
        seq = strip_modifications(glycopeptide)
        for case in self.sequence_to_model:
            if seq in case:
                out.append(self.sequence_to_model[case])
        return out

    def score(self, glycopeptide: FragmentCachingGlycopeptide,
              glycan_composition: Optional[HashableGlycanComposition]=None) -> float:
        if glycan_composition is None:
            glycan_composition = glycopeptide.glycan_composition
        models = self.get_models(glycopeptide)
        if len(models) == 0:
            return MINIMUM
        sites = models[0]
        if len(sites) == 0:
            return MINIMUM
        try:
            acc = []
            for site in sites:
                try:
                    rec = site.glycan_map[glycan_composition]
                    acc.append(rec.score)
                except KeyError:
                    pass
            return max(sum(acc) / len(acc), MINIMUM) if acc else MINIMUM
        except IndexError:
            return MINIMUM

    def __call__(self, glycopeptide):
        return self.get_models(glycopeptide)


class GlycoproteomePriorAnnotator(object):
    target_model: GlycoproteomeModel
    decoy_model: GlycoproteomeModel

    @classmethod
    def load(cls, target_session, decoy_session, fh: io.TextIOBase, hypothesis_id=1, fuzzy=True):
        site_models = GlycosylationSiteModel.load(fh)
        target_model = GlycoproteomeModel.bind_to_hypothesis(
            target_session,
            site_models,
            hypothesis_id=hypothesis_id,
            fuzzy=fuzzy)
        decoy_model = GlycoproteomeModel.bind_to_hypothesis(
            decoy_session,
            site_models,
            hypothesis_id=hypothesis_id,
            fuzzy=fuzzy,
            site_model_cls=ReversedProteinSiteReflectionGlycoproteinSiteSpecificGlycomeModel)
        target_model.stub_proteins()
        decoy_model.stub_proteins()
        return cls(target_model, decoy_model)

    def __init__(self, target_model, decoy_model):
        self.target_model = target_model
        self.decoy_model = decoy_model

    def select_model(self, glycopeptide: FragmentCachingGlycopeptide,
                     structure_type: StructureClassification) -> GlycoproteomeModel:
        if structure_type & StructureClassification.decoy_peptide_target_glycan:
            return self.decoy_model
        else:
            return self.target_model

    def score_glycan(self, glycopeptide: FragmentCachingGlycopeptide,
                     structure_type: StructureClassification,
                     model: GlycoproteomeModel) -> float:
        # Treat decoy glycans identical
        gc = glycopeptide.glycan_composition
        return model.score(glycopeptide, gc)

    def score(self, glycopeptide: FragmentCachingGlycopeptide, structure_type: StructureClassification) -> float:
        model = self.select_model(glycopeptide, structure_type)
        return self.score_glycan(glycopeptide, structure_type, model)
